from tenacity import stop_after_attempt, retry

from kag.common.base.prompt_op import PromptOp
from kag.interface.solver.kag_generator_abc import KAGGeneratorABC
from kag.solver.implementation.default_memory import DefaultMemory


class DefaultGenerator(KAGGeneratorABC):
    """
     The Generator class is an abstract base class for generating responses using a language model module.
     It initializes prompts for judging and generating responses based on the business scene and language settings.
     """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.generate_prompt = PromptOp.load(self.biz_scene, "resp_generator")(
            language=self.language
        )

    @retry(stop=stop_after_attempt(3))
    def generate(self, instruction, memory: DefaultMemory):
        solved_answer = memory.get_solved_answer()
        if solved_answer is not None:
            return solved_answer
        present_memory = memory.serialize_memory()
        return self.llm_module.invoke({'memory': present_memory, 'instruction': instruction}, self.generate_prompt, with_json_parse=False, with_except=True)



import logging
from typing import List

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)


class RespRewriter(PromptOp):
    template_zh = "你是一个智能助手，擅长通过复杂的、多跳的推理帮助用户在多文档中获取信息。请理解当前已知信息与目标问题之间的信息差。" \
                  "你的任务是直接生成一个用于下一步检索的思考问题。" \
                  "不要一次性生成所有思考过程！\n[已知信息]： $memory\n[目标问题]：$instruction\n[你的思考]："
    template_en = "You serve as an intelligent assistant, adept at facilitating users through complex, " \
                  "multi-hop reasoning across multiple documents. Please understand the information gap between the currently known information and the target problem." \
                  "Your task is to generate one thought in the form of question for next retrieval step directly. " \
                  "DON\'T generate the whole thoughts at once!\n[Known information]: $memory\n[Target question]: $instruction\n[You Thought]:"

    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["memory", "instruction"]

    def parse_response_en(self, response: str):
        update_reason_path = []
        split_path = response.split("\n")
        for p in split_path:
            if 'Here are the steps' in p or p == '\n' or p == '':
                continue
            else:
                update_reason_path.append(p)
        logger.debug('cur path:{}'.format(str(update_reason_path)))
        return update_reason_path

    def parse_response_zh(self, response: str):
        update_reason_path = []
        split_path = response.split("\n")
        for p in split_path:
            if '步骤为' in p or p == '\n' or p == '':
                continue
            else:
                update_reason_path.append(p)
        logger.debug('cur path:{}'.format(str(update_reason_path)))
        return update_reason_path

    def parse_response(self, response: str, **kwargs):
        logger.debug('infer result:{}'.format(response))
        if self.language == "en":
            return self.parse_response_en(response)
        return self.parse_response_zh(response)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import re
import sys
import json
from typing import Type,Tuple
import inspect
import os
from pathlib import Path
import importlib
from shutil import copystat, copy2
from typing import Any, Union
from jinja2 import Environment, FileSystemLoader, Template
from stat import S_IWUSR as OWNER_WRITE_PERMISSION


def _register(root, path, files, class_type):
    relative_path = os.path.relpath(path, root)
    module_prefix = relative_path.replace(".", "").replace("/", ".")
    module_prefix = module_prefix + "." if module_prefix else ""
    for file_name in files:
        if file_name.endswith(".py"):
            module_name = module_prefix + os.path.splitext(file_name)[0]
            import importlib

            module = importlib.import_module(module_name)
            classes = inspect.getmembers(module, inspect.isclass)
            for class_name, class_obj in classes:
                if (
                    issubclass(class_obj, class_type)
                    and inspect.getmodule(class_obj) == module
                ):

                    class_type.register(
                        name=class_name,
                        local_path=os.path.join(path, file_name),
                        module_path=module_name,
                    )(class_obj)


def register_from_package(path: str, class_type: Type) -> None:
    """
    Register all classes under the given package.
    Only registered classes can be recognized by kag.
    """
    if not append_python_path(path):
        return
    for root, dirs, files in os.walk(path):
        _register(path, root, files, class_type)
    class_type._has_registered = True


def append_python_path(path: str) -> bool:
    """
    Append the given path to `sys.path`.
    """
    path = Path(path).resolve()
    path = str(path)
    if path not in sys.path:
        sys.path.append(path)
        return True
    return False

def render_template(
    root_dir: Union[str, os.PathLike], file: Union[str, os.PathLike], **kwargs: Any
) -> None:
    env = Environment(loader=FileSystemLoader(root_dir))
    template = env.get_template(str(file))
    content = template.render(kwargs)

    path_obj = Path(root_dir) / file
    render_path = path_obj.with_suffix("") if path_obj.suffix == ".tmpl" else path_obj

    if path_obj.suffix == ".tmpl":
        path_obj.rename(render_path)

    render_path.write_text(content, "utf8")


def copytree(src: Path, dst: Path, **kwargs):
    names = [x.name for x in src.iterdir()]

    if not dst.exists():
        dst.mkdir(parents=True)

    for name in names:
        _name = Template(name).render(**kwargs)
        src_name = src / name
        dst_name = dst / _name
        if src_name.is_dir():
            copytree(src_name, dst_name, **kwargs)
        else:
            copyfile(src_name, dst_name, **kwargs)

    copystat(src, dst)
    _make_writable(dst)


def copyfile(src: Path, dst: Path, **kwargs):
    if dst.exists():
        return
    dst = Path(Template(str(dst)).render(**kwargs))
    copy2(src, dst)
    _make_writable(dst)
    if dst.suffix != ".tmpl":
        return
    render_template('/', dst, **kwargs)


def remove_files_except(path, file, new_file):
    for filename in os.listdir(path):
        file_path = os.path.join(path, filename)
        if os.path.isfile(file_path) and filename != file:
            os.remove(file_path)
    os.rename(path / file, path / new_file)


def _make_writable(path):
    current_permissions = os.stat(path).st_mode
    os.chmod(path, current_permissions | OWNER_WRITE_PERMISSION)


def escape_single_quotes(s: str):
    return s.replace("'", "\\'")


def load_json(content):
    try:
        return json.loads(content)
    except json.JSONDecodeError as e:

        substr = content[: e.colno - 1]
        return json.loads(substr)


def split_module_class_name(name: str, text: str) -> Tuple[str, str]:
    """
    Split `name` as module name and class name pair.

    :param name: fully qualified class name, e.g. ``foo.bar.MyClass``
    :type name: str
    :param text: describe the kind of the class, used in the exception message
    :type text: str
    :rtype: Tuple[str, str]
    :raises RuntimeError: if `name` is not a fully qualified class name
    """
    i = name.rfind(".")
    if i == -1:
        message = "invalid %s class name: %s" % (text, name)
        raise RuntimeError(message)
    module_name = name[:i]
    class_name = name[i + 1 :]
    return module_name, class_name


def dynamic_import_class(name: str, text: str):
    """
    Import the class specified by `name` dyanmically.

    :param name: fully qualified class name, e.g. ``foo.bar.MyClass``
    :type name: str
    :param text: describe the kind of the class, use in the exception message
    :type text: str
    :raises RuntimeError: if `name` is not a fully qualified class name, or
                          the class is not in the module specified by `name`
    :raises ModuleNotFoundError: the module specified by `name` is not found
    """
    module_name, class_name = split_module_class_name(name, text)
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name, None)
    if class_ is None:
        message = "class %r not found in module %r" % (class_name, module_name)
        raise RuntimeError(message)
    if not isinstance(class_, type):
        message = "%r is not a class" % (name,)
        raise RuntimeError(message)
    return class_


def processing_phrases(phrase):
    phrase = str(phrase)
    return re.sub("[^A-Za-z0-9\u4e00-\u9fa5 ]", " ", phrase.lower()).strip()


def to_camel_case(phrase):
    s = processing_phrases(phrase).replace(" ", "_")
    return "".join(
        word.capitalize() if i != 0 else word
        for i, word in enumerate(s.split("_"))
    )


def to_snake_case(name):
    words = re.findall("[A-Za-z][a-z0-9]*", name)
    result = "_".join(words).lower()
    return result



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from abc import ABC
from typing import Type, List, Union

from kag.builder.model.chunk import Chunk
from kag.interface.builder import SplitterABC


class BaseTableSplitter(SplitterABC):
    """
    A base class for splitting table, inheriting from Splitter.
    """

    def split_table(self, org_chunk: Chunk, chunk_size: int = 2000, sep: str = "\n"):
        """
        split markdown format table into smaller markdown table
        """
        try:
            return self._split_table(org_chunk=org_chunk, chunk_size=chunk_size, sep=sep)
        except Exception:
            return None

    def _split_table(self, org_chunk: Chunk, chunk_size: int = 2000, sep: str = "\n"):
        output = []
        content = org_chunk.content
        table_start = content.find("|")
        table_end = content.rfind("|") + 1
        if table_start is None or table_end is None or table_start == table_end:
            return None
        prefix = content[0:table_start].strip("\n ")
        table_rows = content[table_start:table_end].split("\n")
        table_header = table_rows[0]
        table_header_segmentation = table_rows[1]
        suffix = content[table_end:].strip("\n ")

        splitted = []
        cur = [prefix, table_header, table_header_segmentation]
        cur_len = len(prefix)
        for idx, row in enumerate(table_rows[2:]):
            if cur_len > chunk_size:
                cur.append(suffix)
                splitted.append(cur)
                cur_len = 0
                cur = [prefix, table_header, table_header_segmentation]
            cur.append(row)
            cur_len += len(row)

        output = []
        for idx, sentences in enumerate(splitted):
            chunk = Chunk(
                id=f"{org_chunk.id}#{chunk_size}#table#{idx}#LEN",
                name=f"{org_chunk.name}#{idx}",
                content=sep.join(sentences),
                type=org_chunk.type,
                **org_chunk.kwargs
            )
            output.append(chunk)
        return output



from abc import abstractmethod

from kag.interface.solver.kag_memory_abc import KagMemoryABC
from kag.solver.common.base import KagBaseModule


class KagReflectorABC(KagBaseModule):
    def __init__(self, **kwargs):
        """
        Initializes the reflector.
        """
        super().__init__(**kwargs)

    def reflect_query(self, memory: KagMemoryABC, instruction: str) -> (bool, str):
        """
        Reflects on the query and determines whether it can be answered.

        :param memory (KagMemory): The context or memory information to use for rewriting.
        :param instruction (str): The original instruction to be rewritten.
        :return: A tuple (can_answer, refined_query)
            - can_answer: Whether the query can be answered (boolean)
            - refined_query: The refined query (string)
        """
        can_answer = self._can_answer(memory, instruction)
        refined_query = self._refine_query(memory, instruction) if not can_answer else instruction

        return can_answer, refined_query

    @abstractmethod
    def _can_answer(self, memory: KagMemoryABC, instruction: str):
        """
        Determines whether the query can be answered.

        :param memory (KagMemory): The context or memory information to use for rewriting.
        :param instruction (str): The original instruction to be rewritten.
        :return: Whether the query can be answered (boolean)
        """
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    def _refine_query(self, memory: KagMemoryABC, instruction: str):
        """
        Refines the query.

        :param memory (KagMemory): The context or memory information to use for rewriting.
        :param instruction (str): The original instruction to be rewritten.
        :return: The refined query (string)
        """
        raise NotImplementedError("Subclasses must implement this method")



import concurrent.futures
import itertools
import logging
import time
from typing import List

from knext.reasoner import TableResult, ReasonTask
from knext.reasoner.client import ReasonerClient
from kag.solver.logic.core_modules.common.one_hop_graph import copy_one_hop_graph_data, EntityData, Prop, \
    OneHopGraphData, RelationData
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.utils import generate_biz_id_with_type
from kag.solver.logic.core_modules.config import LogicFormConfiguration
from kag.solver.logic.core_modules.parser.logic_node_parser import GetSPONode

logger = logging.getLogger()


class DslRunner:
    def __init__(self, project_id: str, schema: SchemaUtils, config: LogicFormConfiguration):
        # Initialize the DslRunner with project ID, schema, and configuration.
        """
        Initialize the DslRunner for graph database access using Cypher or other languages to retrieve results as OneHopGraph.

        :param project_id: A string representing the ID of the project.
        :param schema: An object of type Schema used for defining the structure of the graph data.
        :param config: An instance of LogicFormConfiguration containing configuration settings.
        """
        self.project_id = project_id
        self.schema = schema
        self.config = config
        self.rc = ReasonerClient(self.config.host_addr, int(self.project_id))

    def get_cached_one_hop_data(self, query_one_graph_cache: dict, biz_id, spo_name):
        if biz_id in query_one_graph_cache.keys():
            return copy_one_hop_graph_data(query_one_graph_cache[biz_id], spo_name)
        return None

    def call_sub_event(self, s_biz_id_set: list, s_node_type: str, o_node_type: str, p_name: str, out_direct: bool,
                       filter_map: dict = None):
        pass

    def run_dsl(self, query, dsl, start_id, params, schema: SchemaUtils, graph_output=False):
        pass

    """
        batch query with s and o
    """

    def query_vertex_one_graph_by_s_o_ids(self, s_node_set: List[EntityData], o_node_set: List[EntityData],
                                          cached_map: dict,n: GetSPONode=None):
        pass

    """
        batch query with s, only get property
    """

    def query_vertex_property_by_s_ids(self, s_biz_id_set: list, s_node_type: str, cached_map: dict):
        pass


class DslRunnerOnGraphStore(DslRunner):

    def __init__(self, project_id, schema: SchemaUtils, config: LogicFormConfiguration):
        super().__init__(project_id, schema, config)
        self.schema = schema

    def run_dsl(self, query, dsl, start_id, params, schema: SchemaUtils, graph_output=False):
        pass

    def _get_filter_gql(self, filter_map: dict, alias: str):
        if filter_map is None or alias not in filter_map:
            return None
        return filter_map[alias]

    def _convert_node_to_json(self, node_str):
        try:
            import json
            node = json.loads(node_str)
        except:
            return {}
        return {
            'id': node['id'],
            'type': node['__label__'],
            'propertyValues': dict(node)
        }

    def _convert_edge_to_json(self, p_str):
        try:
            import json
            p = json.loads(p_str)
        except:
            return {}
        prop = dict(p)
        prop['original_src_id1__'] = p['__from_id__']
        prop['original_dst_id2__'] = p['__to_id__']
        return {
            'type': p['__label__'],
            'propertyValues': prop
        }

    def replace_qota(self, s: str):
        return s.replace("'", "\\'")

    def _generate_gql_type(self, biz_set: list, node_type: str):
        if biz_set is None or len(biz_set) == 0 or node_type is None:
            return ":Entity"
        return f":{node_type}"
    """
        batch query with s and o
    """

    def _do_query_vertex_one_graph_by_s_o_ids(self, s_biz_id: list, s_node_type: str, o_biz_id: list, o_node_type: str,
                                              p_name: str = None, filter_map: dict = None):

        s_biz_id_set = [f'"{self.replace_qota(str(s_id))}"' for s_id in s_biz_id]

        o_biz_id_set = []
        if o_biz_id is not None:
            o_biz_id_set = [f"'{self.replace_qota(str(o_id))}'" for o_id in o_biz_id]

        s_where_cluase = None
        return_cluase = ["p"]
        return_cluase.append("s.id as s_id")
        return_cluase.append("o.id as o_id")

        return_cluase.append("s")

        s_where_filter_cluase = self._get_filter_gql(filter_map, "s")
        if s_where_filter_cluase is not None:
            s_where_cluase = f"{s_where_filter_cluase}"

        o_where_cluase = None

        o_where_filter_cluase = self._get_filter_gql(filter_map, "o")
        if o_where_filter_cluase is not None:
            o_where_cluase = f"{o_where_filter_cluase}"
        return_cluase.append("o")

        p_where_cluase = self._get_filter_gql(filter_map, "p")

        s_gql = f"(s{self._generate_gql_type(s_biz_id_set, s_node_type)})"
        o_gql = f"(o{self._generate_gql_type(o_biz_id_set, o_node_type)})"
        where_cluase = []
        if s_where_cluase:
            where_cluase.append(s_where_cluase)

        if o_where_cluase:
            where_cluase.append(o_where_cluase)

        if p_where_cluase:
            where_cluase.append(p_where_cluase)
        gql_param = {
            "start_alias": "s" if len(s_biz_id_set) > 0 else "o",
            "s_type": s_node_type,
            "o_type": o_node_type
        }
        if len(s_biz_id_set) > 0:
            where_cluase.append(f"s.id in $sid")
            gql_param["sid"] = f'[{",".join(s_biz_id_set)}]'

        if len(o_biz_id_set) > 0:
            where_cluase.append(f"o.id in $oid")
            gql_param["oid"] = f'[{",".join(o_biz_id_set)}]'
        if p_name is None:
            p_name = "rdf_expand()"
        gql_set = self._generate_gql_prio_set(s_node_type, s_biz_id_set, o_node_type, o_biz_id_set, p_name, where_cluase, return_cluase)
        logger.debug("query_vertex_one_graph_by_s_o_ids query " + str(gql_set))

        start_time = time.time()
        for gql in gql_set:
            res = self.rc.syn_execute(gql, **gql_param)
            add_alias = []
            if len(s_biz_id_set) > 0:
                add_alias.append("s")
            if len(o_biz_id_set) > 0:
                add_alias.append("o")
            out = self.parse_one_hot_graph_graph_detail_with_id_map(res.task, add_alias)
            logger.debug(f"query_vertex_one_graph_by_s_o_ids {s_biz_id_set} cost end time {time.time() - start_time}")
            if out is not None and len(out) > 0:
                return out
        return {}

    def _generate_gql_prio_set(self, s_type, s_biz_id_set, o_type, o_biz_id_set, p_type, where_cluase, return_cluase):
        s_gql = f"(s{self._generate_gql_type(s_biz_id_set, s_type)})"
        o_gql = f"(o{self._generate_gql_type(o_biz_id_set, o_type)})"
        rdf_expand_gql = f"""match {s_gql}-[p:rdf_expand()]-{o_gql}
        {'where ' + "and".join(where_cluase) if len(where_cluase) > 0 else ''}
        return {','.join(return_cluase)}"""
        if p_type == "rdf_expand()":
            return [rdf_expand_gql]
        s_without_prefix_type = self.schema.get_label_without_prefix(s_type)
        o_without_prefix_type = self.schema.get_label_without_prefix(o_type)
        ret_gql = []
        if (s_without_prefix_type, o_without_prefix_type) in self.schema.so_p_en and p_type in self.schema.so_p_en[(s_without_prefix_type, o_without_prefix_type)]:
            ret_gql.append(f"""match (s:{s_type})-[p:{p_type}]->(o:{o_type})
        {'where ' + "and".join(where_cluase) if len(where_cluase) > 0 else ''}
        return {','.join(return_cluase)}""")
        if (o_without_prefix_type, s_without_prefix_type) in self.schema.op_s_en  and p_type in self.schema.op_s_en[(o_without_prefix_type, s_without_prefix_type)]:
            ret_gql.append(f"""match (s:{s_type})<-[p:{p_type}]-(o:{o_type})
                    {'where ' + "and".join(where_cluase) if len(where_cluase) > 0 else ''}
                    return {','.join(return_cluase)}""")
        ret_gql.append(rdf_expand_gql)
        return ret_gql

    def _cartesian_product_with_default(self, list1, list2, default=None):
        # 如果任何一个列表为空，使用包含默认值的列表替代
        list1 = list1 if list1 else [default]
        list2 = list2 if list2 else [default]

        return list(itertools.product(list1, list2))

    def _get_node_type_zh(self, node_type):
        if node_type == "attribute":
            return "文本"
        return node_type
    def _get_p_type_name(self, n: GetSPONode):
        if n is None:
            return None
        return n.p.get_entity_first_type()
    def _get_entity_type_name(self, d: EntityData, n: GetSPONode=None, alias=None):
        if d is None and n is None:
            return None
        return self.schema.get_label_within_prefix(n.s.get_entity_first_type() if alias=="s" else n.o.get_entity_first_type()) if d is None else d.type

    def query_vertex_one_graph_by_s_o_ids(self, s_node_set: List[EntityData], o_node_set: List[EntityData],
                                          cached_map: dict, n: GetSPONode=None):
        one_hop_graph_map = {}
        is_enable_cache = True
        if (len(s_node_set) != 0 and len(o_node_set) != 0) or self._get_p_type_name(n):
            # do not cache
            is_enable_cache = False
        s_uncached_biz_id = []
        o_uncached_biz_id = []
        if is_enable_cache:
            for s_node in s_node_set:
                cached_id = generate_biz_id_with_type(s_node.biz_id, self._get_node_type_zh(s_node.type_zh if s_node.type_zh else s_node.type))
                cached_graph = self.get_cached_one_hop_data(cached_map, cached_id, "s")
                if cached_graph:
                    one_hop_graph_map[cached_id] = cached_graph
                else:
                    s_uncached_biz_id.append(s_node)

            for o_node in o_node_set:
                cached_id = generate_biz_id_with_type(o_node.biz_id, self._get_node_type_zh(o_node.type_zh if o_node.type_zh else o_node.type))
                cached_graph = self.get_cached_one_hop_data(cached_map, cached_id, "o")
                if cached_graph:
                    one_hop_graph_map[cached_id] = cached_graph
                else:
                    o_uncached_biz_id.append(o_node)
            if len(s_uncached_biz_id) == 0 and len(o_uncached_biz_id) == 0:
                return one_hop_graph_map
        else:
            s_uncached_biz_id = s_node_set
            o_uncached_biz_id = o_node_set

        combined_list = self._shuffle_query_node(s_uncached_biz_id, o_uncached_biz_id)
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(self._do_query_vertex_one_graph_by_s_o_ids, [] if s_node is None else [s_node.biz_id],
                                self._get_entity_type_name(s_node, n, "s"), [] if o_node is None else [o_node.biz_id],
                                self._get_entity_type_name(o_node, n, "o"), self._get_p_type_name(n)) for
                s_node, o_node in
                combined_list]
            results = [future.result() for future in concurrent.futures.as_completed(futures)]
            for r in results:
                one_hop_graph_map.update(r)
        for node in s_uncached_biz_id + o_uncached_biz_id:
            cached_id = generate_biz_id_with_type(node.biz_id, self._get_node_type_zh(node.type_zh if node.type_zh else node.type))
            if cached_id not in one_hop_graph_map:
                continue
            one_hop_graph = one_hop_graph_map[cached_id]
            one_hop_graph.s.score = node.score
        # updated score
        return one_hop_graph_map

    def _shuffle_query_node(self, s_nodes: List[EntityData], o_nodes: List[EntityData]):
        s_group_types = self._extra_node_id_group_by_type(s_nodes)
        o_group_types = self._extra_node_id_group_by_type(o_nodes)
        combined_list = self._cartesian_product_with_default(s_group_types, o_group_types)
        node_ids = []
        for s_group, o_group in combined_list:
            node_ids = node_ids + self._cartesian_product_with_default(s_group, o_group)
        return node_ids

    def _extra_node_id_group_by_type(self, nodes: List[EntityData]):
        type_map = {}
        for node in nodes:
            if node.type not in type_map:
                type_map[node.type] = []
            type_map[node.type].append(node)
        return list(type_map.values())

    def query_vertex_property_by_s_ids(self, s_biz_id_set: list, s_node_type: str, cached_map: dict):
        one_hop_graph_dict = {}
        s_uncached_biz_id_set = []
        s_node_type_zh = self._get_node_type_zh(s_node_type)
        for s_id in s_biz_id_set:
            cache_id_with_type = generate_biz_id_with_type(s_id, s_node_type_zh)
            cached_graph = self.get_cached_one_hop_data(cached_map, cache_id_with_type, "s")
            if cached_graph:
                one_hop_graph_dict[cache_id_with_type] = cached_graph
            else:
                s_uncached_biz_id_set.append(s_id)
        if len(s_uncached_biz_id_set) == 0:
            return one_hop_graph_dict

        s_biz_id_set = s_uncached_biz_id_set
        s_biz_id_set = [f"'{self.replace_qota(str(s_id))}'" for s_id in s_biz_id_set]

        return_cluase = []
        id_rep = "id"

        gql_param = {
            "start_alias": "s"
        }
        if len(s_biz_id_set) > 0:
            s_where_cluase = "s.id in $sid"
            gql_param["sid"] = f"[{','.join(s_biz_id_set)}]"
        else:
            return {}

        return_cluase.append("s")
        return_cluase.append("s.id as s_id")

        gql = f"""
        match (s{':' + 'Entity' if len(s_biz_id_set) != 0 else ''}{' where ' + s_where_cluase if s_where_cluase else ''})
        return {','.join(return_cluase)}
        """
        logger.debug("query_vertex_one_graph_by_s_o_ids query " + gql)
        start_time = time.time()

        res = self.rc.syn_execute(gql, **gql_param)
        add_alias = []
        if len(s_biz_id_set) > 0:
            add_alias.append("s")
        out = self.parse_one_hot_graph_graph_detail_with_id_map(res.task, add_alias)
        one_hop_graph_dict.update(out)
        logger.debug(f"query_vertex_one_graph_by_s_o_ids {s_biz_id_set} cost end time {time.time() - start_time}")
        return one_hop_graph_dict

    def _trans_normal_p_json(self, p_json, s_json, o_json):
        s_type = s_json['type']
        s_biz_id = s_json["propertyValues"]["id"]
        o_type = o_json['type']
        o_biz_id = o_json["propertyValues"]["id"]
        p_total_type_name = p_json['type']
        if len(s_type) > len(o_type):
            p_type = p_json['type'].replace(s_type, "").replace(o_type, "").replace("_", "")
        else:
            p_type = p_json['type'].replace(o_type, "").replace(s_type, "").replace("_", "")
        p_info = {}
        from_id = None
        to_id = None
        for property_key in p_json['propertyValues'].keys():
            if property_key == "original_src_id1__":
                from_id = p_json['propertyValues'][property_key]
            elif property_key == "original_dst_id2__":
                to_id = p_json['propertyValues'][property_key]
            else:
                p_info[property_key] = p_json['propertyValues'][property_key]
        if from_id is None or to_id is None:
            return None
        """
        rel.from_id = json_dict["__from_id__"]
        rel.from_type = json_dict["__from_id_type__"]
        rel.end_id = json_dict["__to_id__"]
        rel.end_type = json_dict["__to_id_type__"]
        rel.type = json_dict["__label__"]
        """
        if s_type in p_total_type_name and s_type != o_type:
            is_out_edge = p_total_type_name.startswith(s_type)
        else:
            is_out_edge = from_id == s_biz_id

        if is_out_edge:
            p_info.update({
                "__label__": p_type,
                "__from_id__": s_biz_id,
                "__from_id_type__": s_type,
                "__to_id__": o_biz_id,
                "__to_id_type__": o_type
            })
        else:
            p_info.update({
                "__label__": p_type,
                "__from_id__": o_biz_id,
                "__from_id_type__": o_type,
                "__to_id__": s_biz_id,
                "__to_id_type__": s_type
            })

        return p_info

    def _check_need_property(self, json_data):
        if "propertyValues" not in json_data.keys():
            return False
        if "type" not in json_data.keys():
            return False
        if "id" not in json_data.keys():
            return False
        if "id" not in json_data["propertyValues"].keys():
            return False
        return True

    def parse_one_hot_graph_graph_detail_with_id_map(self, task_resp: ReasonTask, add_alias: list):
        one_hop_graph_map = {}
        if task_resp is None or task_resp.status != "FINISH":
            return one_hop_graph_map
        detail = task_resp.result_table_result
        if detail.total == 0:
            return one_hop_graph_map
        tmp_graph_parse_result_map = {}
        s_index = -1
        p_index = -1
        o_index = -1
        # format header s, p
        for i in range(len(detail.header)):
            if detail.header[i] == "s":
                s_index = i
            elif detail.header[i] == "p":
                p_index = i
            elif detail.header[i] == "o":
                o_index = i
        if p_index is None:
            return one_hop_graph_map

        # get all EntityData first
        for data in detail.rows:
            s_entity = None
            o_entity = None
            s_json = {}
            if s_index != -1:
                s_json = self._convert_node_to_json(data[s_index])
                if self._check_need_property(s_json) is False:
                    continue
                prop_values = s_json['propertyValues']
                s_biz_id = prop_values["id"]
                s_type_name = s_json["type"]
                s_biz_id_with_type_name = generate_biz_id_with_type(s_biz_id, self._get_node_type_zh(s_type_name))
                if s_biz_id_with_type_name not in tmp_graph_parse_result_map.keys():
                    s_entity = EntityData()
                    s_entity.type = s_type_name
                    s_entity.type_zh = self._get_node_type_zh(s_type_name)
                    s_entity.prop = Prop.from_dict(prop_values, s_entity.type, None)
                    s_entity.biz_id = s_biz_id
                    s_entity.name = prop_values["name"]
                    if "description" in prop_values.keys():
                        s_entity.description = prop_values["description"]
                    one_hop_graph = OneHopGraphData(None, "s")
                    one_hop_graph.s = s_entity
                    tmp_graph_parse_result_map[s_biz_id_with_type_name] = one_hop_graph
                else:
                    s_entity = tmp_graph_parse_result_map[s_biz_id_with_type_name].s

                if "s" in add_alias and s_biz_id_with_type_name not in one_hop_graph_map.keys():
                    one_hop_graph_map[s_biz_id_with_type_name] = tmp_graph_parse_result_map[s_biz_id_with_type_name]
            else:
                s_biz_id_with_type_name = None

            if o_index == -1 or p_index == -1:
                continue

            o_json = {}
            if o_index != -1:
                o_json = self._convert_node_to_json(data[o_index])
                if self._check_need_property(o_json) is False:
                    continue
                prop_values = o_json['propertyValues']
                o_biz_id = prop_values["id"]
                o_type_name = o_json["type"]
                o_biz_id_with_type_name = generate_biz_id_with_type(o_biz_id, self._get_node_type_zh(o_type_name))
                if o_biz_id_with_type_name not in tmp_graph_parse_result_map.keys():
                    o_entity = EntityData()
                    o_entity.type = o_type_name
                    o_entity.type_zh = self._get_node_type_zh(o_type_name)
                    o_entity.prop = Prop.from_dict(prop_values, o_entity.type, None)
                    o_entity.biz_id = o_biz_id

                    o_entity.name = prop_values["name"]
                    if "description" in o_json.keys():
                        o_entity.description = prop_values["description"]
                    one_hop_graph = OneHopGraphData(None, "o")
                    one_hop_graph.s = o_entity
                    tmp_graph_parse_result_map[o_biz_id_with_type_name] = one_hop_graph
                else:
                    o_entity = tmp_graph_parse_result_map[o_biz_id_with_type_name].s

                if "o" in add_alias and o_biz_id_with_type_name not in one_hop_graph_map.keys():
                    one_hop_graph_map[o_biz_id_with_type_name] = tmp_graph_parse_result_map[o_biz_id_with_type_name]
            else:
                o_biz_id_with_type_name = None

            if s_entity is None and o_entity is None:
                logger.info("parse_one_hot_graph_graph_detail_with_id_map entity is None")
                continue

            if p_index == -1:
                continue
            p_json = self._convert_edge_to_json(data[p_index])
            p_json = self._trans_normal_p_json(p_json, s_json, o_json)
            rel = RelationData.from_dict(p_json, None)
            # if rel.type in ['similarity', 'source']:
            #     continue
            if s_entity is None:
                s_entity = EntityData()
                if rel.from_id == o_entity.biz_id:
                    s_entity.biz_id = rel.end_id
                    s_entity.type = rel.end_type
                    s_entity.type_zh = self._get_node_type_zh(rel.end_type)
                else:
                    s_entity.biz_id = rel.from_id
                    s_entity.type = rel.from_type
                    s_entity.type_zh = self._get_node_type_zh(rel.from_type)

            if o_entity is None:
                o_entity = EntityData()
                if rel.from_id == s_entity.biz_id:
                    o_entity.biz_id = rel.end_id
                    o_entity.type = rel.end_type
                    o_entity.type_zh = self._get_node_type_zh(rel.end_type)
                else:
                    o_entity.biz_id = rel.from_id
                    o_entity.type = rel.from_type
                    o_entity.type_zh = self._get_node_type_zh(rel.from_type)

            if generate_biz_id_with_type(rel.from_id, rel.from_type) == generate_biz_id_with_type(s_entity.biz_id,
                                                                                                  s_type_name):
                rel.from_entity = s_entity
                rel.end_entity = o_entity

                if s_biz_id_with_type_name is not None and s_biz_id_with_type_name in tmp_graph_parse_result_map.keys():
                    if rel.type in tmp_graph_parse_result_map[s_biz_id_with_type_name].out_relations.keys():
                        tmp_graph_parse_result_map[s_biz_id_with_type_name].out_relations[rel.type].append(rel)
                    else:
                        tmp_graph_parse_result_map[s_biz_id_with_type_name].out_relations[rel.type] = [rel]
                if o_biz_id_with_type_name is not None and o_biz_id_with_type_name in tmp_graph_parse_result_map.keys():
                    if rel.type in tmp_graph_parse_result_map[o_biz_id_with_type_name].in_relations.keys():
                        tmp_graph_parse_result_map[o_biz_id_with_type_name].in_relations[rel.type].append(rel)
                    else:
                        tmp_graph_parse_result_map[o_biz_id_with_type_name].in_relations[rel.type] = [rel]
            else:
                rel.from_entity = o_entity
                rel.from_alias = "o"
                rel.end_entity = s_entity
                rel.end_alias = "s"
                if s_biz_id_with_type_name is not None and s_biz_id_with_type_name in tmp_graph_parse_result_map.keys():
                    if rel.type in tmp_graph_parse_result_map[s_biz_id_with_type_name].in_relations.keys():
                        tmp_graph_parse_result_map[s_biz_id_with_type_name].in_relations[rel.type].append(rel)
                    else:
                        tmp_graph_parse_result_map[s_biz_id_with_type_name].in_relations[rel.type] = [rel]

                if o_biz_id_with_type_name is not None and o_biz_id_with_type_name in tmp_graph_parse_result_map.keys():
                    if rel.type in tmp_graph_parse_result_map[o_biz_id_with_type_name].out_relations.keys():
                        tmp_graph_parse_result_map[o_biz_id_with_type_name].out_relations[rel.type].append(rel)
                    else:
                        tmp_graph_parse_result_map[o_biz_id_with_type_name].out_relations[rel.type] = [rel]
        return one_hop_graph_map



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import logging
import os
import re
import threading
import time
from abc import ABCMeta

import schedule
from neo4j import GraphDatabase

from kag.common.graphstore.graph_store import GraphStore
from kag.common.utils import escape_single_quotes
from knext.schema.model.base import IndexTypeEnum

logger = logging.getLogger(__name__)

class SingletonMeta(ABCMeta):
    """
    Thread-safe Singleton metaclass
    """
    _instances = {}
    _lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        uri = kwargs.get('uri')
        user = kwargs.get('user')
        password = kwargs.get('password')
        database = kwargs.get('database', 'neo4j')
        key = (cls, uri, user, password, database)

        with cls._lock:
            if key not in cls._instances:
                cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]


class Neo4jClient(GraphStore, metaclass=SingletonMeta):

    def __init__(self, uri, user, password, database="neo4j", init_type="write", interval_minutes=10):
        self._driver = GraphDatabase.driver(uri, auth=(user, password))
        logger.info(f"init Neo4jClient uri: {uri} database: {database}")
        self._database = database
        self._lucene_special_chars = "\\+-!():^[]\"{}~*?|&/"
        self._lucene_pattern = self._get_lucene_pattern()
        self._simple_ident = "[A-Za-z_][A-Za-z0-9_]*"
        self._simple_ident_pattern = re.compile(self._simple_ident)
        self._vec_meta = dict()
        self._vec_meta_ts = 0.0
        self._vec_meta_timeout = 60.0
        self._vectorizer = None
        self._allGraph = "allGraph_0"
        if init_type == "write":
            self._labels = self._create_unique_constraint()
            self._create_all_graph(self._allGraph)
        self.schedule_constraint(interval_minutes)
        # self.create_text_index(["Chunk"], ["content"])
        self.refresh_vector_index_meta(force=True)

    def close(self):
        self._driver.close()

    def schedule_constraint(self, interval_minutes):

        def job():
            try:
                self._labels = self._create_unique_constraint()
                self._update_pagerank_graph()
            except Exception as e:
                import traceback
                logger.error(f"Error run scheduled job: {traceback.format_exc()}")

        def run_scheduled_tasks():
            while True:
                schedule.run_pending()
                time.sleep(1)

        if interval_minutes > 0:
            schedule.every(interval_minutes).minutes.do(job)
            scheduler_thread = threading.Thread(target=run_scheduled_tasks, daemon=True)
            scheduler_thread.start()

    def get_all_entity_labels(self):
        with self._driver.session(database=self._database) as session:
            result = session.run("CALL db.labels()")
            labels = [record[0] for record in result]
            return labels

    def run_script(self, script):
        with self._driver.session(database=self._database) as session:
            return list(session.run(script))

    def _create_unique_constraint(self):
        with self._driver.session(database=self._database) as session:
            result = session.run("CALL db.labels()")
            labels = [record[0] for record in result if record[0] != "Entity"]
            for label in labels:
                self._create_unique_index_constraint(self, label, session)
            return labels

    @staticmethod
    def _create_unique_index_constraint(self, label, session):
        constraint_name = f"uniqueness_{label}_id"
        create_constraint_query = f"CREATE CONSTRAINT {self._escape_neo4j(constraint_name)} IF NOT EXISTS FOR (n:{self._escape_neo4j(label)}) REQUIRE n.id IS UNIQUE"

        try:
            result = session.run(create_constraint_query)
            result.consume()
            logger.debug(f"Unique constraint created for constraint_name: {constraint_name}")
        except Exception as e:
            logger.debug(f"warn creating constraint for {constraint_name}: {e}")
            self._create_index_constraint(self, label, session)

    @staticmethod
    def _create_index_constraint(self, label, session):
        index_name = f"index_{label}_id"
        create_constraint_query = f"CREATE INDEX {self._escape_neo4j(index_name)} IF NOT EXISTS FOR (n:{self._escape_neo4j(label)}) ON (n.id)"
        try:
            result = session.run(create_constraint_query)
            result.consume()
            logger.debug(f"index constraint created for constraint_name: {index_name}")
        except Exception as e:
            logger.warn(f"warn creating index constraint for {index_name}: {e}")

    def _update_pagerank_graph(self):
        all_graph_0 = "allGraph_0"
        all_graph_1 = "allGraph_1"

        if self._allGraph == all_graph_0:
            all_graph = all_graph_1
        else:
            all_graph = all_graph_0

        logger.debug(f"update pagerank graph for {all_graph}")
        self._create_all_graph(all_graph)

        logger.debug(f"drop old pagerank graph for {self._allGraph}")
        self._drop_all_graph(self._allGraph)
        self._allGraph = all_graph

    def create_pagerank_graph(self):
        self._drop_all_graph(self._allGraph)
        self._create_all_graph(self._allGraph)

    def initialize_schema(self, schema_types):
        for spg_type in schema_types:
            label = spg_type
            properties = schema_types[spg_type].properties
            if properties:
                for property_key in properties:
                    if property_key == "name":
                        self.create_vector_index(label, property_key)
                    index_type = properties[property_key].index_type
                    if index_type:
                        if index_type == IndexTypeEnum.Text:
                            pass
                        elif index_type == IndexTypeEnum.Vector:
                            self.create_vector_index(label, property_key)
                        elif index_type == IndexTypeEnum.TextAndVector:
                            self.create_vector_index(label, property_key)
                        else:
                            logger.info(f"Undefined IndexTypeEnum {index_type}")
        labels, property_keys = self._collect_text_index_info(schema_types)
        self.create_text_index(labels, property_keys)
        self.create_vector_index("Entity", "name")
        self.create_vector_index("Entity", "desc")
        self.refresh_vector_index_meta(force=True)

    def _collect_text_index_info(self, schema_types):
        labels = {}
        property_keys = {}
        for spg_type in schema_types:
            label = spg_type
            properties = schema_types[spg_type].properties
            if properties:
                label_property_keys = {}
                for property_key in properties:
                    index_type = properties[property_key].index_type
                    if property_key == "name" or index_type and index_type in (IndexTypeEnum.Text, IndexTypeEnum.TextAndVector):
                        label_property_keys[property_key] = True
                if label_property_keys:
                    labels[label] = True
                    property_keys.update(label_property_keys)
        return tuple(labels.keys()), tuple(property_keys.keys())

    def upsert_node(self, label, properties, id_key="id", extra_labels=("Entity",)):
        self._preprocess_node_properties(label, properties, extra_labels)
        with self._driver.session(database=self._database) as session:
            if label not in self._labels:
                self._create_unique_index_constraint(self, label, session)
            try:
                return session.execute_write(self._upsert_node, self, label, id_key, properties, extra_labels)
            except Exception as e:
                logger.error(f"upsert_node label:{label} properties:{properties} Exception: {e}")
                return None

    @staticmethod
    def _upsert_node(tx, self, label, id_key, properties, extra_labels):
        if not label:
            logger.warning("label cannot be None or empty strings")
            return None
        query = (f"MERGE (n:{self._escape_neo4j(label)} {{{self._escape_neo4j(id_key)}: $properties.{self._escape_neo4j(id_key)}}}) "
                  "SET n += $properties ")
        if extra_labels:
            query += f", n:{':'.join(self._escape_neo4j(extra_label) for extra_label in extra_labels)} "
        query += "RETURN n"
        result = tx.run(query, properties=properties)
        return result.single()[0]

    def upsert_nodes(self, label, properties_list, id_key="id", extra_labels=("Entity",)):
        self._preprocess_node_properties_list(label, properties_list, extra_labels)
        with self._driver.session(database=self._database) as session:
            if label not in self._labels:
                self._create_unique_index_constraint(self, label, session)
            try:
                return session.execute_write(self._upsert_nodes, self, label, properties_list, id_key, extra_labels)
            except Exception as e:
                logger.error(f"upsert_nodes label:{label} properties:{properties_list} Exception: {e}")
                return None

    @staticmethod
    def _upsert_nodes(tx, self, label, properties_list, id_key, extra_labels):
        if not label:
            logger.warning("label cannot be None or empty strings")
            return None
        query = ("UNWIND $properties_list AS properties "
                f"MERGE (n:{self._escape_neo4j(label)} {{{self._escape_neo4j(id_key)}: properties.{self._escape_neo4j(id_key)}}}) "
                 "SET n += properties ")
        if extra_labels:
            query += f", n:{':'.join(self._escape_neo4j(extra_label) for extra_label in extra_labels)} "
        query += "RETURN n"
        result = tx.run(query, properties_list=properties_list)
        return [record['n'] for record in result]

    def _get_embedding_vector(self, properties, vector_field):
        for property_key, property_value in properties.items():
            field_name = self._create_vector_field_name(property_key)
            if field_name != vector_field:
                continue
            if not property_value:
                return None
            if not isinstance(property_value, str):
                message = f"property {property_key!r} must be string to generate embedding vector"
                raise RuntimeError(message)
            try:
                vector = self.vectorizer.vectorize(property_value)
                return vector
            except Exception as e:
                logger.info(f"An error occurred while vectorizing property {property_key!r}: {e}")
            return None
        return None

    def _preprocess_node_properties(self, label, properties, extra_labels):
        if self._vectorizer is None:
            return
        self.refresh_vector_index_meta()
        vec_meta = self._vec_meta
        labels = [label]
        if extra_labels:
            labels.extend(extra_labels)
        for label in labels:
            if label not in vec_meta:
                continue
            for vector_field in vec_meta[label]:
                if vector_field in properties:
                    continue
                embedding_vector = self._get_embedding_vector(properties, vector_field)
                if embedding_vector is not None:
                    properties[vector_field] = embedding_vector

    def _preprocess_node_properties_list(self, label, properties_list, extra_labels):
        for properties in properties_list:
            self._preprocess_node_properties(label, properties, extra_labels)

    def batch_preprocess_node_properties(self, node_batch, extra_labels=("Entity",)):
        if self._vectorizer is None:
            return

        class EmbeddingVectorPlaceholder(object):
            def __init__(self, number, properties, vector_field, property_key, property_value):
                self._number = number
                self._properties = properties
                self._vector_field = vector_field
                self._property_key = property_key
                self._property_value = property_value
                self._embedding_vector = None

            def replace(self):
                if self._embedding_vector is not None:
                    self._properties[self._vector_field] = self._embedding_vector

            def __repr__(self):
                return repr(self._number)

        class EmbeddingVectorManager(object):
            def __init__(self):
                self._placeholders = []

            def get_placeholder(self, graph_store, properties, vector_field):
                for property_key, property_value in properties.items():
                    field_name = graph_store._create_vector_field_name(property_key)
                    if field_name != vector_field:
                        continue
                    if not property_value:
                        return None
                    if not isinstance(property_value, str):
                        message = f"property {property_key!r} must be string to generate embedding vector"
                        raise RuntimeError(message)
                    num = len(self._placeholders)
                    placeholder = EmbeddingVectorPlaceholder(num, properties, vector_field, property_key, property_value)
                    self._placeholders.append(placeholder)
                    return placeholder
                return None

            def _get_text_batch(self):
                text_batch = dict()
                for placeholder in self._placeholders:
                    property_value = placeholder._property_value
                    if property_value not in text_batch:
                        text_batch[property_value] = list()
                    text_batch[property_value].append(placeholder)
                return text_batch

            def _generate_vectors(self, vectorizer, text_batch):
                texts = list(text_batch)
                vectors = vectorizer.vectorize(texts)
                return vectors

            def _fill_vectors(self, vectors, text_batch):
                for vector, (_text, placeholders) in zip(vectors, text_batch.items()):
                    for placeholder in placeholders:
                        placeholder._embedding_vector = vector

            def batch_vectorize(self, vectorizer):
                text_batch = self._get_text_batch()
                vectors = self._generate_vectors(vectorizer, text_batch)
                self._fill_vectors(vectors, text_batch)

            def patch(self):
                for placeholder in self._placeholders:
                    placeholder.replace()

        manager = EmbeddingVectorManager()
        self.refresh_vector_index_meta()
        vec_meta = self._vec_meta
        for node_item in node_batch:
            label, properties = node_item
            labels = [label]
            if extra_labels:
                labels.extend(extra_labels)
            for label in labels:
                if label not in vec_meta:
                    continue
                for vector_field in vec_meta[label]:
                    if vector_field in properties:
                        continue
                    placeholder = manager.get_placeholder(self, properties, vector_field)
                    if placeholder is not None:
                        properties[vector_field] = placeholder
        manager.batch_vectorize(self._vectorizer)
        manager.patch()

    def get_node(self, label, id_value, id_key="id"):
        with self._driver.session(database=self._database) as session:
            return session.execute_read(self._get_node, self, label, id_key, id_value)

    @staticmethod
    def _get_node(tx, self, label, id_key, id_value):
        query = f"MATCH (n:{self._escape_neo4j(label)} {{{self._escape_neo4j(id_key)}: $id_value}}) RETURN n"
        result = tx.run(query, id_value=id_value)
        single_result = result.single()
        # print(f"single_result: {single_result}")
        if single_result is not None:
            return single_result[0]
        else:
            return None

    def delete_node(self, label, id_value, id_key="id"):
        with self._driver.session(database=self._database) as session:
            try:
                session.execute_write(self._delete_node, self, label, id_key, id_value)
            except Exception as e:
                logger.error(f"delete_node label:{label} Exception: {e}")

    @staticmethod
    def _delete_node(tx, self, label, id_key, id_value):
        query = f"MATCH (n:{self._escape_neo4j(label)} {{{self._escape_neo4j(id_key)}: $id_value}}) DETACH DELETE n"
        tx.run(query, id_value=id_value)

    def delete_nodes(self, label, id_values, id_key="id"):
        with self._driver.session(database=self._database) as session:
            session.execute_write(self._delete_nodes, self, label, id_key, id_values)

    @staticmethod
    def _delete_nodes(tx, self, label, id_key, id_values):
        query = f"UNWIND $id_values AS id_value MATCH (n:{self._escape_neo4j(label)} {{{self._escape_neo4j(id_key)}: id_value}}) DETACH DELETE n"
        tx.run(query, id_values=id_values)

    def upsert_relationship(self, start_node_label, start_node_id_value,
                            end_node_label, end_node_id_value, rel_type,
                            properties, upsert_nodes=True, start_node_id_key="id", end_node_id_key="id"):
        rel_type = self._escape_neo4j(rel_type)
        with self._driver.session(database=self._database) as session:
            try:
                return session.execute_write(self._upsert_relationship, self, start_node_label, start_node_id_key,
                                         start_node_id_value, end_node_label, end_node_id_key,
                                         end_node_id_value, rel_type, properties, upsert_nodes)
            except Exception as e:
                logger.error(f"upsert_relationship rel_type:{rel_type} properties:{properties} Exception: {e}")
                return None

    @staticmethod
    def _upsert_relationship(tx, self, start_node_label, start_node_id_key, start_node_id_value,
                             end_node_label, end_node_id_key, end_node_id_value,
                             rel_type, properties, upsert_nodes):
        if not start_node_label or not end_node_label or not rel_type:
            logger.warning("start_node_label, end_node_label, and rel_type cannot be None or empty strings")
            return None
        if upsert_nodes:
            query = (
                f"MERGE (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: $start_node_id_value}}) "
                f"MERGE (b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: $end_node_id_value}}) "
                f"MERGE (a)-[r:{self._escape_neo4j(rel_type)}]->(b) SET r += $properties RETURN r"
            )
        else:
            query = (
                f"MATCH (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: $start_node_id_value}}), "
                f"(b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: $end_node_id_value}}) "
                f"MERGE (a)-[r:{self._escape_neo4j(rel_type)}]->(b) SET r += $properties RETURN r"
            )
        result = tx.run(query, start_node_id_value=start_node_id_value,
                        end_node_id_value=end_node_id_value, properties=properties)
        return result.single()

    def upsert_relationships(self, start_node_label, end_node_label, rel_type, relations,
                             upsert_nodes=True, start_node_id_key="id", end_node_id_key="id"):
        with self._driver.session(database=self._database) as session:
            try:
                return session.execute_write(self._upsert_relationships, self, relations, start_node_label,
                                         start_node_id_key, end_node_label, end_node_id_key, rel_type, upsert_nodes)
            except Exception as e:
                logger.error(f"upsert_relationships rel_type:{rel_type} relations:{relations} Exception: {e}")
                return None

    @staticmethod
    def _upsert_relationships(tx, self, relations, start_node_label, start_node_id_key,
                              end_node_label, end_node_id_key, rel_type, upsert_nodes):
        if not start_node_label or not end_node_label or not rel_type:
            logger.warning("start_node_label, end_node_label, and rel_type cannot be None or empty strings")
            return None
        if upsert_nodes:
            query = (
                "UNWIND $relations AS relationship "
                f"MERGE (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: relationship.start_node_id}}) "
                f"MERGE (b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: relationship.end_node_id}}) "
                f"MERGE (a)-[r:{self._escape_neo4j(rel_type)}]->(b) SET r += relationship.properties RETURN r"
            )
        else:
            query = (
                "UNWIND $relations AS relationship "
                f"MATCH (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: relationship.start_node_id}}) "
                f"MATCH (b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: relationship.end_node_id}}) "
                f"MERGE (a)-[r:{self._escape_neo4j(rel_type)}]->(b) SET r += relationship.properties RETURN r"
            )

        result = tx.run(query, relations=relations,
                        start_node_label=start_node_label, start_node_id_key=start_node_id_key,
                        end_node_label=end_node_label, end_node_id_key=end_node_id_key,
                        rel_type=rel_type)
        return [record['r'] for record in result]

    def delete_relationship(self, start_node_label, start_node_id_value,
                            end_node_label, end_node_id_value, rel_type,
                            start_node_id_key="id", end_node_id_key="id"):
        with self._driver.session(database=self._database) as session:
            try:
                session.execute_write(self._delete_relationship, self, start_node_label, start_node_id_key,
                                      start_node_id_value, end_node_label, end_node_id_key,
                                      end_node_id_value, rel_type)
            except Exception as e:
                logger.error(f"delete_relationship rel_type:{rel_type} Exception: {e}")


    @staticmethod
    def _delete_relationship(tx, self, start_node_label, start_node_id_key, start_node_id_value,
                             end_node_label, end_node_id_key, end_node_id_value, rel_type):
        query = (
            f"MATCH (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: $start_node_id_value}})-[r:{self._escape_neo4j(rel_type)}]->"
            f"(b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: $end_node_id_value}}) DELETE r"
        )
        tx.run(query, start_node_id_value=start_node_id_value, end_node_id_value=end_node_id_value)

    def delete_relationships(self, start_node_label, start_node_id_values,
                             end_node_label, end_node_id_values, rel_type,
                             start_node_id_key="id", end_node_id_key="id"):
        with self._driver.session(database=self._database) as session:
            session.execute_write(self._delete_relationships, self,
                                  start_node_label, start_node_id_key, start_node_id_values,
                                  end_node_label, end_node_id_key, end_node_id_values, rel_type)

    @staticmethod
    def _delete_relationships(tx, self, start_node_label, start_node_id_key, start_node_id_values,
                              end_node_label, end_node_id_key, end_node_id_values, rel_type):
        query = (
            "UNWIND $start_node_id_values AS start_node_id_value "
            "UNWIND $end_node_id_values AS end_node_id_value "
            f"MATCH (a:{self._escape_neo4j(start_node_label)} {{{self._escape_neo4j(start_node_id_key)}: start_node_id_value}})-[r:{self._escape_neo4j(rel_type)}]->"
            f"(b:{self._escape_neo4j(end_node_label)} {{{self._escape_neo4j(end_node_id_key)}: end_node_id_value}}) DELETE r"
        )
        tx.run(query, start_node_id_values=start_node_id_values, end_node_id_values=end_node_id_values)

    def _get_lucene_pattern(self):
        string = re.escape(self._lucene_special_chars)
        pattern = "([" + string + "])"
        pattern = re.compile(pattern)
        return pattern

    def _escape_lucene(self, string):
        result = self._lucene_pattern.sub(r"\\\1", string)
        return result

    def _make_lucene_query(self, string):
        string = self._escape_lucene(string)
        result = string.lower()
        return result

    def _get_utf16_codepoints(self, string):
        result = []
        for ch in string:
            data = ch.encode("utf-16-le")
            for i in range(0, len(data), 2):
                value = int.from_bytes(data[i:i+2], "little")
                result.append(value)
        return tuple(result)

    def _escape_neo4j(self, name):
        match = self._simple_ident_pattern.fullmatch(name)
        if match is not None:
            return name
        string = "`"
        for ch in name:
            if ch == "`":
                string += "``"
            elif ch.isascii() and ch.isprintable():
                string += ch
            else:
                values = self._get_utf16_codepoints(ch)
                for value in values:
                    string += "\\u%04X" % value
        string += "`"
        return string

    def _to_snake_case(self, name):
        import re
        words = re.findall("[A-Za-z][a-z0-9]*", name)
        result = "_".join(words).lower()
        return result

    def _create_vector_index_name(self, label, property_key):
        name = f"{label}_{property_key}_vector_index"
        name = self._to_snake_case(name)
        return "_" + name

    def _create_vector_field_name(self, property_key):
        name = f"{property_key}_vector"
        name = self._to_snake_case(name)
        return "_" + name

    def create_index(self, label, property_key, index_name=None):
        with self._driver.session(database=self._database) as session:
            session.execute_write(self._create_index, self, label, property_key, index_name)

    @staticmethod
    def _create_index(tx, self, label, property_key, index_name):
        if not label or not property_key:
            return
        if index_name is None:
            query = f"CREATE INDEX IF NOT EXISTS FOR (n:{self._escape_neo4j(label)}) ON (n.{self._escape_neo4j(property_key)})"
        else:
            query = f"CREATE INDEX {self._escape_neo4j(index_name)} IF NOT EXISTS FOR (n:{self._escape_neo4j(label)}) ON (n.{self._escape_neo4j(property_key)})"
        tx.run(query)

    def create_text_index(self, labels, property_keys, index_name=None):
        if not labels or not property_keys:
            return
        if index_name is None:
            index_name = "_default_text_index"
        label_spec = "|".join(self._escape_neo4j(label) for label in labels)
        property_spec = ", ".join(f"n.{self._escape_neo4j(key)}" for key in property_keys)
        query = (
            f"CREATE FULLTEXT INDEX {self._escape_neo4j(index_name)} IF NOT EXISTS "
            f"FOR (n:{label_spec}) ON EACH [{property_spec}]"
        )
        def do_create_text_index(tx):
            tx.run(query)
        with self._driver.session(database=self._database) as session:
            session.execute_write(do_create_text_index)
        return index_name

    def create_vector_index(self, label, property_key, index_name=None,
                            vector_dimensions=768, metric_type="cosine",
                            hnsw_m=None, hnsw_ef_construction=None):
        if index_name is None:
            index_name = self._create_vector_index_name(label, property_key)
        if not property_key.lower().endswith("vector"):
            property_key = self._create_vector_field_name(property_key)
        with self._driver.session(database=self._database) as session:
            session.execute_write(self._create_vector_index, self, label, property_key, index_name,
                                  vector_dimensions, metric_type, hnsw_m, hnsw_ef_construction)
        self.refresh_vector_index_meta(force=True)
        return index_name

    @staticmethod
    def _create_vector_index(tx, self, label, property_key, index_name, vector_dimensions, metric_type, hnsw_m, hnsw_ef_construction):
        query = (
            f"CREATE VECTOR INDEX {self._escape_neo4j(index_name)} IF NOT EXISTS FOR (n:{self._escape_neo4j(label)}) ON (n.{self._escape_neo4j(property_key)}) "
             "OPTIONS { indexConfig: {"
             "  `vector.dimensions`: $vector_dimensions,"
             "  `vector.similarity_function`: $metric_type"
        )
        if hnsw_m is not None:
            query += ",  `vector.hnsw.m`: $hnsw_m"
        if hnsw_ef_construction is not None:
            query += ",  `vector.hnsw.ef_construction`: $hnsw_ef_construction"
        query += "}}"
        tx.run(query, vector_dimensions=vector_dimensions, metric_type=metric_type,
               hnsw_m=hnsw_m, hnsw_ef_construction=hnsw_ef_construction)

    def refresh_vector_index_meta(self, force=False):
        import time
        if not force and time.time() - self._vec_meta_ts < self._vec_meta_timeout:
            return
        def do_refresh_vector_index_meta(tx):
            query = "SHOW VECTOR INDEX"
            res = tx.run(query)
            data = res.data()
            meta = dict()
            for record in data:
                if record["entityType"] == "NODE":
                    label, = record["labelsOrTypes"]
                    vector_field, = record["properties"]
                    if vector_field.startswith("_") and vector_field.endswith("_vector"):
                        if label not in meta:
                            meta[label] = []
                        meta[label].append(vector_field)
            self._vec_meta = meta
            self._vec_meta_ts = time.time()
        with self._driver.session(database=self._database) as session:
            session.execute_read(do_refresh_vector_index_meta)

    def delete_index(self, index_name):
        with self._driver.session(database=self._database) as session:
            session.execute_write(self._delete_index, self, index_name)

    @staticmethod
    def _delete_index(tx, self, index_name):
        query = f"DROP INDEX {self._escape_neo4j(index_name)} IF EXISTS"
        tx.run(query)

    @property
    def vectorizer(self):
        if self._vectorizer is None:
            message = "vectorizer is not initialized"
            raise RuntimeError(message)
        return self._vectorizer

    @vectorizer.setter
    def vectorizer(self, value):
        self._vectorizer = value

    def text_search(self, query_string, label_constraints=None, topk=10, index_name=None):
        if index_name is None:
            index_name = "_default_text_index"
        if label_constraints is None:
            pass
        elif isinstance(label_constraints, str):
            label_constraints = self._escape_neo4j(label_constraints)
        elif isinstance(label_constraints, (list, tuple)):
            label_constraints = "|".join(self._escape_neo4j(label_constraint) for label_constraint in label_constraints)
        else:
            message = f"invalid label_constraints: {label_constraints!r}"
            raise RuntimeError(message)
        if label_constraints is None:
            query = ("CALL db.index.fulltext.queryNodes($index_name, $query_string) "
                     "YIELD node AS node, score "
                     "RETURN node, score")
        else:
            query = ("CALL db.index.fulltext.queryNodes($index_name, $query_string) "
                     "YIELD node AS node, score "
                    f"WHERE (node:{label_constraints}) "
                     "RETURN node, score")
        query += " LIMIT $topk"
        query_string = self._make_lucene_query(query_string)

        def do_text_search(tx):
            res = tx.run(query, query_string=query_string, topk=topk, index_name=index_name)
            data = res.data()
            return data

        with self._driver.session(database=self._database) as session:
            return session.execute_read(do_text_search)

    def vector_search(self, label, property_key, query_text_or_vector, topk=10, index_name=None, ef_search=None):
        if ef_search is not None:
            if ef_search < topk:
                message = f"ef_search must be greater than or equal to topk; {ef_search!r} is invalid"
                raise ValueError(message)
        self.refresh_vector_index_meta()
        if index_name is None:
            vec_meta = self._vec_meta
            if label not in vec_meta:
                logger.warning(f"vector index not defined for label, return empty. label: {label}, "
                               f"property_key: {property_key}, query_text_or_vector: {query_text_or_vector}.")
                return []
            vector_field = self._create_vector_field_name(property_key)
            if vector_field not in vec_meta[label]:
                logger.warning(f"vector index not defined for field, return empty. label: {label}, "
                               f"property_key: {property_key}, query_text_or_vector: {query_text_or_vector}.")
                return []
        if index_name is None:
            index_name = self._create_vector_index_name(label, property_key)
        if isinstance(query_text_or_vector, str):
            query_vector = self.vectorizer.vectorize(query_text_or_vector)
        else:
            query_vector = query_text_or_vector

        def do_vector_search(tx):
            if ef_search is not None:
                query = ("CALL db.index.vector.queryNodes($index_name, $ef_search, $query_vector) "
                         "YIELD node, score "
                         "RETURN node, score, labels(node) as __labels__"
                        f"LIMIT {topk}")
                res = tx.run(query, query_vector=query_vector, ef_search=ef_search, index_name=index_name)
            else:
                query = ("CALL db.index.vector.queryNodes($index_name, $topk, $query_vector) "
                         "YIELD node, score "
                         "RETURN node, score, labels(node) as __labels__")
                res = tx.run(query, query_vector=query_vector, topk=topk, index_name=index_name)
            data = res.data()
            for record in data:
                record["node"]["__labels__"] = record["__labels__"]
                del record["__labels__"]
            return data

        with self._driver.session(database=self._database) as session:
            return session.execute_read(do_vector_search)

    def _create_all_graph(self, graph_name):
        with self._driver.session(database=self._database) as session:
            logger.debug(f"create pagerank graph graph_name：{graph_name} database：{self._database}")
            result = session.run(f"""
            CALL gds.graph.exists('{graph_name}') YIELD exists
            WHERE exists
            CALL gds.graph.drop('{graph_name}') YIELD graphName
            RETURN graphName
            """)
            summary = result.consume()
            logger.debug(f"create pagerank graph exists graph_name：{graph_name} database：{self._database} succeed "
                  f"executed：{summary.result_available_after} consumed：{summary.result_consumed_after}")

            result = session.run(f"""
            CALL gds.graph.project('{graph_name}','*','*')
            YIELD graphName, nodeCount AS nodes, relationshipCount AS rels
            RETURN graphName, nodes, rels
            """)
            summary = result.consume()
            logger.debug(f"create pagerank graph graph_name：{graph_name} database：{self._database} succeed "
                  f"executed：{summary.result_available_after} consumed：{summary.result_consumed_after}")

    def _drop_all_graph(self, graph_name):
        with self._driver.session(database=self._database) as session:
            logger.debug(f"drop pagerank graph graph_name：{graph_name} database：{self._database}")
            result = session.run(f"""
            CALL gds.graph.exists('{graph_name}') YIELD exists
            WHERE exists
            CALL gds.graph.drop('{graph_name}') YIELD graphName
            RETURN graphName
            """)
            result.consume()
            logger.debug(f"drop pagerank graph graph_name：{graph_name} database：{self._database} succeed")

    def execute_pagerank(self, iterations=20, damping_factor=0.85):
        with self._driver.session(database=self._database) as session:
            return session.execute_write(self._execute_pagerank, iterations, damping_factor)

    @staticmethod
    def _execute_pagerank(tx, iterations, damping_factor):
        query = (
            "CALL algo.pageRank.stream("
            "{iterations: $iterations, dampingFactor: $damping_factor}) "
            "YIELD nodeId, score "
            "RETURN algo.getNodeById(nodeId) AS node, score "
            "ORDER BY score DESC"
        )
        result = tx.run(query, iterations=iterations, damping_factor=damping_factor)
        return [{"node": record["node"], "score": record["score"]} for record in result]

    def get_pagerank_scores(self, start_nodes, target_type):
        with self._driver.session(database=self._database) as session:
            all_graph = self._allGraph
            self._exists_all_graph(session, all_graph)
            data = session.execute_write(self._get_pagerank_scores, self, all_graph, start_nodes, target_type)
        return data

    @staticmethod
    def _get_pagerank_scores(tx, self, graph_name, start_nodes, return_type):
        match_clauses = []
        match_identify = []
        for index, node in enumerate(start_nodes):
            node_type, node_name = node['type'], node['name']
            node_identify = f"node_{index}"
            match_clauses.append(f"MATCH ({node_identify}:{self._escape_neo4j(node_type)} {{name: '{escape_single_quotes(node_name)}'}})")
            match_identify.append(node_identify)

        match_query = ' '.join(match_clauses)
        match_identify_str = ', '.join(match_identify)

        pagerank_query = f"""
        {match_query}
        CALL gds.pageRank.stream('{graph_name}',{{
            maxIterations: 20,
            dampingFactor: 0.85,
            sourceNodes: [{match_identify_str}]
        }})
        YIELD nodeId, score
        MATCH (m:{return_type}) WHERE id(m) = nodeId
        RETURN id(m) AS g_id, gds.util.asNode(nodeId).id AS id, score
        ORDER BY score DESC
        """

        result = tx.run(pagerank_query)
        return [{"id": record["id"], "score": record["score"]} for record in result]

    @staticmethod
    def _exists_all_graph(session, graph_name):
        try:
            logger.debug(f"exists pagerank graph graph_name：{graph_name}")
            result = session.run(f"""
            CALL gds.graph.exists('{graph_name}') YIELD exists
            WHERE NOT exists
            CALL gds.graph.project('{graph_name}','*','*')
            YIELD graphName, nodeCount AS nodes, relationshipCount AS rels
            RETURN graphName, nodes, rels
            """)
            summary = result.consume()
            logger.debug(f"exists pagerank graph graph_name：{graph_name} succeed "
                  f"executed：{summary.result_available_after} consumed：{summary.result_consumed_after}")
        except Exception as e:
            logger.debug(f"Error exists pagerank graph {graph_name}: {e}")

    def count(self, label):
        with self._driver.session(database=self._database) as session:
            return session.execute_read(self._count, self, label)

    @staticmethod
    def _count(tx, self, label):
        query = f"MATCH (n:{self._escape_neo4j(label)}) RETURN count(n)"
        result = tx.run(query)
        single_result = result.single()
        if single_result is not None:
            return single_result[0]

    def create_database(self, database):
        with self._driver.session(database=self._database) as session:
            database = database.lower()
            result = session.run(f"CREATE DATABASE {self._escape_neo4j(database)} IF NOT EXISTS")
            summary = result.consume()
            logger.info(f"create_database {database} succeed "
                  f"executed：{summary.result_available_after} consumed：{summary.result_consumed_after}")

    def delete_all_data(self, database):
        if self._database != database:
            raise ValueError(f"Error: Current database ({self._database}) is not the same as the target database ({database}).")

        with self._driver.session(database=database) as session:
            while True:
                result = session.run("MATCH (n)  WITH n LIMIT 100000  DETACH DELETE n RETURN count(*)")
                count = result.single()[0]
                logger.info(f"Deleted {count} nodes in this batch.")
                if count == 0:
                    logger.info("All data has been deleted.")
                    break

    def run_cypher_query(self, database, query, parameters=None):
        if database and self._database != database:
            raise ValueError(f"Current database ({self._database}) is not the same as the target database ({database}).")

        with self._driver.session(database=database) as session:
            result = session.run(query, parameters)
            return [record for record in result]



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os
from typing import List, Type, Dict

import pandas as pd

from kag.builder.model.chunk import Chunk
from kag.interface.builder.reader_abc import SourceReaderABC
from knext.common.base.runnable import Input, Output


class CSVReader(SourceReaderABC):
    """
    A class for reading CSV files, inheriting from `SourceReader`.
    Supports converting CSV data into either a list of dictionaries or a list of Chunk objects.

    Args:
        output_type (Output): Specifies the output type, which can be "Dict" or "Chunk".
        **kwargs: Additional keyword arguments passed to the parent class constructor.
    """

    def __init__(self, output_type="Chunk", **kwargs):
        super().__init__(**kwargs)
        if output_type == "Dict":
            self.output_types = Dict[str, str]
        else:
            self.output_types = Chunk
        self.id_col = kwargs.get("id_col", "id")
        self.name_col = kwargs.get("name_col", "name")
        self.content_col = kwargs.get("content_col", "content")

    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return self._output_types

    @output_types.setter
    def output_types(self, output_types):
        self._output_types = output_types

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Reads a CSV file and converts the data format based on the output type.

        Args:
            input (Input): Input parameter, expected to be a string representing the path to the CSV file.
            **kwargs: Additional keyword arguments, which may include `id_column`, `name_column`, `content_column`, etc.

        Returns:
            List[Output]:
                - If `output_types` is `Chunk`, returns a list of Chunk objects.
                - If `output_types` is `Dict`, returns a list of dictionaries.
        """

        try:
            data = pd.read_csv(input)
            data = data.astype(str)
        except Exception as e:
            raise IOError(f"Failed to read the file: {e}")

        if self.output_types == Chunk:
            chunks = []
            basename, _ = os.path.splitext(os.path.basename(input))
            for idx, row in enumerate(data.to_dict(orient="records")):
                kwargs = {k: v for k, v in row.items() if k not in [self.id_col, self.name_col, self.content_col]}
                chunks.append(
                    Chunk(
                        id=row.get(self.id_col) or Chunk.generate_hash_id(f"{input}#{idx}"),
                        name=row.get(self.name_col) or f"{basename}#{idx}",
                        content=row[self.content_col],
                        **kwargs
                    )
                )
            return chunks
        else:
            return data.to_dict(orient="records")



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
import os
from typing import List, Type, Dict, Union

from kag.builder.component.reader.markdown_reader import MarkDownReader
from kag.builder.model.chunk import Chunk
from kag.interface.builder.reader_abc import SourceReaderABC
from knext.common.base.runnable import Input, Output

from kag.common.llm.client import LLMClient


class JSONReader(SourceReaderABC):
    """
    A class for reading JSON files, inheriting from `SourceReader`.
    Supports converting JSON data into either a list of dictionaries or a list of Chunk objects.

    Args:
        output_types (Output): Specifies the output type, which can be "Dict" or "Chunk".
        **kwargs: Additional keyword arguments passed to the parent class constructor.
    """

    def __init__(self, output_type="Chunk", **kwargs):
        super().__init__(**kwargs)
        if output_type == "Dict":
            self.output_types = Dict[str, str]
        else:
            self.output_types = Chunk
        self.id_col = kwargs.get("id_col", "id")
        self.name_col = kwargs.get("name_col", "name")
        self.content_col = kwargs.get("content_col", "content")

    @property
    def input_types(self) -> Type[Input]:
        return str

    @property
    def output_types(self) -> Type[Output]:
        return self._output_types

    @output_types.setter
    def output_types(self, output_types):
        self._output_types = output_types

    @staticmethod
    def _read_from_file(file_path: str) -> Union[dict, list]:
        """
        Safely reads JSON from a file and returns its content.

        Args:
            file_path (str): The path to the JSON file.

        Returns:
            Union[dict, list]: The parsed JSON content.

        Raises:
            ValueError: If there is an error reading the JSON file.
        """
        try:
            with open(file_path, "r") as file:
                return json.load(file)
        except json.JSONDecodeError as e:
            raise ValueError(f"Error reading JSON from file: {e}")
        except FileNotFoundError as e:
            raise ValueError(f"File not found: {e}")

    @staticmethod
    def _parse_json_string(json_string: str) -> Union[dict, list]:
        """
        Parses a JSON string and returns its content.

        Args:
            json_string (str): The JSON string to parse.

        Returns:
            Union[dict, list]: The parsed JSON content.

        Raises:
            ValueError: If there is an error parsing the JSON string.
        """
        try:
            return json.loads(json_string)
        except json.JSONDecodeError as e:
            raise ValueError(f"Error parsing JSON string: {e}")

    def invoke(self, input: str, **kwargs) -> List[Output]:
        """
        Parses the input string data and generates a list of Chunk objects or returns the original data.

        This method supports receiving JSON-formatted strings. It extracts specific fields based on provided keyword arguments.
        It can read from a file or directly parse a string. If the input data is in the expected format, it generates a list of Chunk objects;
        otherwise, it throws a ValueError if the input is not a JSON array or object.

        Args:
            input (str): The input data, which can be a JSON string or a file path.
            **kwargs: Keyword arguments used to specify the field names for ID, name, and content.

        Returns:
            List[Output]: A list of Chunk objects or the original data.

        Raises:
            ValueError: If the input data format is incorrect or parsing fails.
        """

        id_col = kwargs.get("id_col", "id")
        name_col = kwargs.get("name_col", "name")
        content_col = kwargs.get("content_col", "content")
        self.id_col = id_col
        self.name_col = name_col
        self.content_col = content_col
        try:
            if os.path.exists(input):
                corpus = self._read_from_file(input)
            else:
                corpus = self._parse_json_string(input)
        except ValueError as e:
            raise e

        if not isinstance(corpus, (list, dict)):
            raise ValueError("Expected input to be a JSON array or object")

        if isinstance(corpus, dict):
            corpus = [corpus]

        if self.output_types == Chunk:
            chunks = []
            basename, _ = os.path.splitext(os.path.basename(input))
            for idx, item in enumerate(corpus):
                if not isinstance(item, dict):
                    continue

                chunk = Chunk(
                    id=item.get(self.id_col) or Chunk.generate_hash_id(f"{input}#{idx}"),
                    name=item.get(self.name_col) or f"{basename}#{idx}",
                    content=item.get(self.content_col),
                )
                chunks.append(chunk)

            return chunks
        else:
            return corpus

if __name__ == "__main__":
    reader = JSONReader()
    json_string = '''[
            {
                "title": "test_json", 
                "text": "Test content"
            }
        ]'''
    chunks = reader.invoke(json_string,name_column="title",content_col = "text")
    res = 1


import logging
import time
from typing import List

from kag.interface.retriever.kg_retriever_abc import KGRetrieverABC
from kag.solver.logic.core_modules.common.base_model import SPOEntity, LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph, EntityData, RelationData, \
    OneHopGraphData
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.common.text_sim_by_vector import TextSimilarity
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor
from kag.solver.logic.core_modules.parser.logic_node_parser import GetSPONode
from kag.solver.logic.core_modules.retriver.entity_linker import EntityLinkerBase, spo_entity_linker
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner

logger = logging.getLogger()


class GetSPOExecutor(OpExecutor):
    """
    Executor for the 'get_spo' operator.

    This class is used to retrieve one-hop graphs based on the given parameters.
    It extends the base `OpExecutor` class and initializes additional components specific to retrieving SPO triples.
    """
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, retrieval_spo: KGRetrieverABC,
                 el: EntityLinkerBase,
                 dsl_runner: DslRunner, query_one_graph_cache: dict, debug_info: dict, text_similarity: TextSimilarity=None,**kwargs):
        """
        Initializes the GetSPOExecutor with necessary components.

        Parameters:
            nl_query (str): Natural language query string.
            kg_graph (KgGraph): Knowledge graph object for subsequent queries and parsing.
            schema (SchemaUtils): Semantic structure definition to assist in the parsing process.
            retrieval_spo (KGRetrieverABC): Retrieval object for SPO triples.
            el (EntityLinkerBase): Entity linker for entity linking tasks.
            dsl_runner (DslRunner): Runner cypher for query graph database.
            query_one_graph_cache (dict): Cache for storing results of one-hop graph queries.
            debug_info (dict): Debug information dictionary to record debugging information during parsing.
        """
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)
        self.retrieval_spo = retrieval_spo
        self.dsl_runner = dsl_runner
        self.query_one_graph_cache = query_one_graph_cache
        self.el = el

        self.text_similarity = text_similarity or TextSimilarity()

    def _find_relation_result(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData], req_id: str):
        one_kg_graph = KgGraph()
        is_find_relation = False
        if n.p.get_entity_first_type_or_zh() is None and n.o.get_entity_first_type_or_zh() is None:
            is_find_relation = True
            for one_hop_graph in one_hop_graph_list:
                rel_set = one_hop_graph.get_all_relation_value()
                one_kg_graph_ = KgGraph()
                recall_alias_name = n.s.alias_name if one_hop_graph.s_alias_name == "s" else n.o.alias_name
                one_kg_graph_.entity_map[recall_alias_name] = [one_hop_graph.s]
                one_kg_graph_.edge_map[n.p.alias_name] = rel_set
                one_kg_graph.merge_kg_graph(one_kg_graph_)
        return one_kg_graph, is_find_relation

    def _get_spo_value_in_one_hop_graph_set(self, n: GetSPONode, one_hop_graph_list: List[OneHopGraphData], req_id: str):
        process_kg, is_rel = self._find_relation_result(n, one_hop_graph_list, req_id)
        if is_rel:
            return process_kg
        return self.retrieval_spo.retrieval_relation(n, one_hop_graph_list, req_id=req_id, debug_info = self.debug_info)

    def _run_query_vertex_one_graph(self, s_node_set: List[EntityData], o_node_set: List[EntityData], n: GetSPONode=None):
        return self.dsl_runner.query_vertex_one_graph_by_s_o_ids(s_node_set,
                                                                 o_node_set,
                                                                 self.query_one_graph_cache, n)

    def _execute_get_spo_by_set(self, n: GetSPONode, s_node_set: List[EntityData], o_node_set: List[EntityData], req_id):
        start_time = time.time()
        kg_graph = KgGraph()
        kg_graph.query_graph[n.p.alias_name] = {
            "s": n.s.alias_name,
            "p": n.p.alias_name,
            "o": n.o.alias_name
        }

        if (s_node_set is None or len(s_node_set) == 0) and (o_node_set is None or len(o_node_set) == 0):
            logger.info(f"{req_id} not found id is spo " + str(n))
            return kg_graph
        one_hop_graph_map = self._run_query_vertex_one_graph(s_node_set, o_node_set, n)
        end_time = time.time()
        logger.debug(f"{req_id} execute_get_spo_by_set {n} recall subgraph cost {end_time - start_time}")
        if len(one_hop_graph_map) == 0:
            logger.debug(f"{req_id} execute_get_spo_by_set one_hop_graph_map is empty")
            return kg_graph
        kg_graph.nodes_alias.append(n.s.alias_name)
        kg_graph.nodes_alias.append(n.o.alias_name)
        kg_graph.edge_alias.append(n.p.alias_name)

        start_time = time.time()
        one_hop_graph_list = []
        for biz_id in one_hop_graph_map.keys():
            self.query_one_graph_cache[biz_id] = one_hop_graph_map[biz_id]
            one_hop_graph_list.append(one_hop_graph_map[biz_id])

        res = self._get_spo_value_in_one_hop_graph_set(n, one_hop_graph_list, req_id)
        kg_graph.merge_kg_graph(res)

        logger.debug(
            f"{req_id} execute_get_spo_by_set merged kg graph ={kg_graph.to_edge_str()} cost = {time.time() - start_time}")
        return kg_graph

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> KgGraph:
        kg_graph = KgGraph()
        if not isinstance(logic_node, GetSPONode):
            return kg_graph
        n = logic_node

        self.kg_graph.logic_form_base[n.s.alias_name] = n.s
        self.kg_graph.logic_form_base[n.p.alias_name] = n.p
        self.kg_graph.logic_form_base[n.o.alias_name] = n.o

        # 实体标准化, 针对有实体名称的节点，需要做链指
        el_results, el_request, err_msg, call_result_data = spo_entity_linker(self.kg_graph,
                                                                              n,
                                                                              self.nl_query,
                                                                              self.el,
                                                                              self.schema,
                                                                              req_id,
                                                                              param)
        if el_request and el_request['entity_mentions']:
            self.debug_info['el'] = self.debug_info['el'] + el_results
            self.debug_info['el_detail'] = self.debug_info['el_detail'] + [{
                "el_request": el_request,
                'el_results': el_results,
                'el_debug_result': call_result_data,
                'err_msg': err_msg
            }]
            n.to_std(n.args)

        s_data_set = []
        relation_data_set = []

        if isinstance(n.s, SPOEntity) and len(n.s.id_set) > 0:
            s_data_set = self._get_entity_node_from_lf(n.s)
        else:
            s_data_set_up = self.kg_graph.get_entity_by_alias(n.s.alias_name)
            if s_data_set_up is not None:
                for s_data in s_data_set_up:
                    if isinstance(s_data, EntityData) and s_data.type != "attribute":
                        s_data_set.append(s_data)
                    if isinstance(s_data, RelationData):
                        relation_data_set.append(s_data)

        if len(relation_data_set) > 0:
            logger.info(f"{req_id} get_spo relation_data_set is not empty {str(relation_data_set)}, need get prop")
            return kg_graph

        o_data_set = []
        if n.o:
            if isinstance(n.o, SPOEntity) and len(n.o.id_set) > 0:
                o_data_set = self._get_entity_node_from_lf(n.o)
            else:
                o_data_set_up = self.kg_graph.get_entity_by_alias(n.o.alias_name)
                if o_data_set_up is not None:
                    for o_data in o_data_set_up:
                        o_data_set.append(o_data)
        cur_spo_graph = None
        if s_data_set:
            cur_spo_graph = self._execute_get_spo_by_set(n, s_data_set, [], req_id)
        if o_data_set:
            cur_spo_graph_o = self._execute_get_spo_by_set(n, [], o_data_set, req_id)
            if cur_spo_graph is None:
                cur_spo_graph = cur_spo_graph_o
            else:
                cur_spo_graph.merge_kg_graph(cur_spo_graph_o)
        kg_graph = cur_spo_graph
        return kg_graph

    def _get_entity_node_from_lf(self, e: SPOEntity):
        if not e.id_set:
            return []
        ret = []
        for biz_id in e.id_set:
            d = EntityData()
            d.biz_id = biz_id
            d.type = e.get_entity_first_type()
            d.type_zh = e.get_entity_first_type_or_en()
            ret.append(d)
        return ret


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import pprint
from typing import Dict, List, Any

from knext.schema.client import BASIC_TYPES
from kag.builder.model.spg_record import SPGRecord
from knext.schema.model.base import BaseSpgType


class Node(object):
    id: str
    name: str
    label: str
    properties: Dict[str, str]
    hash_map: Dict[int, str] = dict()

    def __init__(self, _id: str, name: str, label: str, properties: Dict[str, str]):
        self.name = name
        self.label = label
        self.properties = properties
        self.id = _id

    @classmethod
    def from_spg_record(cls, idx, spg_record: SPGRecord):
        return cls(
            _id=idx,
            name=spg_record.get_property("name"),
            label=spg_record.spg_type_name,
            properties=spg_record.properties,
        )

    @staticmethod
    def unique_key(spg_record):
        return spg_record.spg_type_name + '_' + spg_record.get_property("name", "")

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "label": self.label,
            "properties": self.properties,
        }

    @classmethod
    def from_dict(cls, input: Dict):
        return cls(
            _id=input["id"],
            name=input["name"],
            label=input["label"],
            properties=input["properties"],
        )

    def __eq__(self, other):
        return self.name == other.name and self.label == other.label and self.properties == other.properties


class Edge(object):
    id: str
    from_id: str
    from_type: str
    to_id: str
    to_type: str
    label: str
    properties: Dict[str, str]

    def __init__(
            self, _id: str, from_node: Node, to_node: Node, label: str, properties: Dict[str, str]
    ):
        self.from_id = from_node.id
        self.from_type = from_node.label
        self.to_id = to_node.id
        self.to_type = to_node.label
        self.label = label
        self.properties = properties
        if not _id:
            _id = id(self)
        self.id = _id

    @classmethod
    def from_spg_record(
            cls, s_idx, subject_record: SPGRecord, o_idx, object_record: SPGRecord, label: str
    ):
        from_node = Node.from_spg_record(s_idx, subject_record)
        to_node = Node.from_spg_record(o_idx, object_record)

        return cls(_id="", from_node=from_node, to_node=to_node, label=label, properties={})

    def to_dict(self):
        return {
            "id": self.id,
            "from": self.from_id,
            "to": self.to_id,
            "fromType": self.from_type,
            "toType": self.to_type,
            "label": self.label,
            "properties": self.properties,
        }

    @classmethod
    def from_dict(cls, input: Dict):
        return cls(
            _id=input["id"],
            from_node=Node(_id=input["from"], name=input["from"],label=input["fromType"], properties={}),
            to_node=Node(_id=input["to"], name=input["to"], label=input["toType"], properties={}),
            label=input["label"],
            properties=input["properties"],
        )

    def __eq__(self, other):
        return self.from_id == other.from_id and self.to_id == other.to_id and self.label == other.label and self.properties == other.properties and self.from_type == other.from_type and self.to_type == other.to_type


class SubGraph(object):
    id: str
    nodes: List[Node] = list()
    edges: List[Edge] = list()

    def __init__(self, nodes: List[Node], edges: List[Edge]):
        self.nodes = nodes
        self.edges = edges

    def add_node(self, id: str, name: str, label: str, properties=None):
        if not properties:
            properties = dict()
        self.nodes.append(Node(_id=id, name=name, label=label, properties=properties))
        return self

    def add_edge(self, s_id: str, s_label: str, p: str, o_id: str, o_label: str, properties=None):
        if not properties:
            properties = dict()
        s_node = Node(_id=s_id, name=s_id, label=s_label, properties={})
        o_node = Node(_id=o_id, name=o_id, label=o_label, properties={})
        self.edges.append(Edge(_id="", from_node=s_node, to_node=o_node, label=p, properties=properties))
        return self

    def to_dict(self):
        return {
            "resultNodes": [n.to_dict() for n in self.nodes],
            "resultEdges": [e.to_dict() for e in self.edges],
        }

    def __repr__(self):
        return pprint.pformat(self.to_dict())

    def merge(self, sub_graph: 'SubGraph'):
        self.nodes.extend(sub_graph.nodes)
        self.edges.extend(sub_graph.edges)

    @classmethod
    def from_spg_record(
        cls, spg_types: Dict[str, BaseSpgType], spg_records: List[SPGRecord]
    ):
        sub_graph = cls([], [])
        for record in spg_records:
            s_id = record.id
            s_name = record.name
            s_label = record.spg_type_name.split('.')[-1]
            properties = record.properties

            spg_type = spg_types.get(record.spg_type_name)
            for prop_name, prop_value in record.properties.items():
                if prop_name in spg_type.properties:
                    from knext.schema.model.property import Property
                    prop: Property = spg_type.properties.get(prop_name)
                    o_label = prop.object_type_name.split('.')[-1]
                    if o_label not in BASIC_TYPES:
                        prop_value_list = prop_value.split(',')
                        for o_id in prop_value_list:
                            sub_graph.add_edge(s_id=s_id, s_label=s_label, p=prop_name, o_id=o_id, o_label=o_label)
                        properties.pop(prop_name)
            sub_graph.add_node(id=s_id, name=s_name, label=s_label, properties=properties)

        return sub_graph

    @classmethod
    def from_dict(cls, input: Dict[str, Any]):
        return cls(
            nodes=[Node.from_dict(node) for node in input["resultNodes"]],
            edges=[Edge.from_dict(edge) for edge in input["resultEdges"]],
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import logging
import os
import re
from typing import List, Type

from kag.interface.builder import SplitterABC
from kag.builder.prompt.semantic_seg_prompt import SemanticSegPrompt
from kag.builder.model.chunk import Chunk
from knext.common.base.runnable import Input, Output
from kag.common.llm.client.llm_client import LLMClient

logger = logging.getLogger(__name__)


class SemanticSplitter(SplitterABC):
    """
    A class for semantically splitting text into smaller chunks based on the content's structure and meaning.
    Inherits from the Splitter class.

    Attributes:
        kept_char_pattern (re.Pattern): Regex pattern to match Chinese/ASCII characters.
        split_length (int): The maximum length of each chunk after splitting.
        llm_client (LLMClient): Instance of LLMClient initialized with `model` config.
        semantic_seg_op (SemanticSegPrompt): Instance of SemanticSegPrompt for semantic segmentation.
    """

    def __init__(self, split_length: int = 1000, **kwargs):
        super().__init__(**kwargs)
        # Chinese/ASCII characters
        self.kept_char_pattern = re.compile(
            r"[^\u4e00-\u9fa5\u3000-\u303F\uFF01-\uFF0F\uFF1A-\uFF20\uFF3B-\uFF40\uFF5B-\uFF65\x00-\x7F]+"
        )
        self.split_length = int(split_length)
        self.llm = self._init_llm()
        language = os.getenv("KAG_PROMPT_LANGUAGE", "zh")
        self.semantic_seg_op = SemanticSegPrompt(language)

    @property
    def input_types(self) -> Type[Input]:
        return Chunk

    @property
    def output_types(self) -> Type[Output]:
        return Chunk

    @staticmethod
    def parse_llm_output(content: str, llm_output: list):
        """
        Parses the output from the LLM to generate segmented information.

        Args:
            content (str): The original content being split.
            llm_output (list): Output from the LLM indicating segment locations and abstracts.

        Returns:
            list: A list of dictionaries containing segment names, contents, and lengths.
        """
        seg_info = llm_output

        seg_info.sort()
        locs = [x[0] for x in seg_info]
        abstracts = [x[1] for x in seg_info]
        locs.append(len(content))
        splitted = []
        for idx in range(len(abstracts)):
            start = locs[idx]
            end = locs[idx + 1]
            splitted.append(
                {
                    "name": abstracts[idx],
                    "content": content[start:end],
                    "length": end - start,
                }
            )

        return splitted

    def semantic_chunk(
        self,
        org_chunk: Chunk,
        chunk_size: int = 1000,
    ) -> List[Chunk]:
        """
        Splits the given content into semantic chunks using an LLM.

        Args:
            org_chunk (Chunk): The original chunk to be split.
            chunk_size (int, optional): Maximum size of each chunk. Defaults to 1000.

        Returns:
            List[Chunk]: A list of Chunk objects representing the split content.
        """
        result = self.llm.invoke({"input": org_chunk.content}, self.semantic_seg_op)
        splitted = self.parse_llm_output(org_chunk.content, result)
        logger.debug(f"splitted = {splitted}")
        chunks = []
        for idx, item in enumerate(splitted):
            split_name = item["name"]
            if len(item["content"]) < chunk_size:
                chunk = Chunk(
                    id=f"{org_chunk.id}#{chunk_size}#{idx}#SEM",
                    name=f"{org_chunk.name}#{split_name}",
                    content=item["content"],
                    abstract=item["name"],
                    **org_chunk.kwargs
                )
                chunks.append(chunk)
            else:
                print("chunk over size")
                innerChunk = Chunk(
                    id=Chunk.generate_hash_id(item["content"]),
                    name=f"{org_chunk.name}#{split_name}",
                    content=item["content"],
                )
                chunks.extend(
                    self.semantic_chunk(
                        innerChunk, chunk_size
                    )
                )
        return chunks

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Invokes the splitting process on the provided input.

        Args:
            input (Input): The input to be processed.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list of outputs generated from the input.
        """
        chunks = self.semantic_chunk(input, self.split_length)
        return chunks



# coding=utf8
import os
import sys
from typing import List

from kag.solver.implementation.default_kg_retrieval import KGRetrieverByLlm
from kag.solver.logic.core_modules.common.base_model import SPOEntity
from kag.solver.logic.core_modules.common.one_hop_graph import EntityData

sys.path.append('../logic_form_executor/')
current_dir = os.path.dirname(os.path.abspath(__file__))
import logging

logger = logging.getLogger()


class SchemaRetrieval(KGRetrieverByLlm):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def retrieval_entity(self, mention_entity: SPOEntity, topk=1, **kwargs) -> List[EntityData]:
        # 根据mention召回
        label = self.schema.get_label_within_prefix('SemanticConcept')
        typed_nodes = self.sc.search_vector(
            label=label, property_key="name", query_vector=self.vectorizer.vectorize(mention_entity.entity_name), topk=1
        )
        recalled_entity = EntityData()
        recalled_entity.type = "SemanticConcept"
        if len(typed_nodes) == 0 or typed_nodes[0]["score"] < 0.9:
            recalled_entity = EntityData()
            recalled_entity.biz_id = "Entity"
            recalled_entity.name = "Entity"
            return [recalled_entity]
        recalled_entity.biz_id = typed_nodes[0]["node"]["name"]
        recalled_entity.name = typed_nodes[0]["node"]["name"]
        recalled_entity.score = typed_nodes[0]["score"]
        return [recalled_entity]



import logging
from typing import List

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)


class SpoRetrieval(PromptOp):
    template_zh = """示例 1:
问题：Woman's Viewpoint 是一本英国出版物吗？
SPO 提及：出版物[Woman's Viewpoint] 国籍 国家
SPO 候选项：['Woman's Viewpoint 从 1923 年出版到 1927 年', 'Woman's Viewpoint 由 Florence M. Sterling 出版', 'Woman's Viewpoint 创立于 1923 年', 'Woman's Viewpoint 在德克萨斯州创立', 'Woman's Viewpoint 是一本女性杂志', 'Rolandos Liatsos 出演 Woman in Mind', 'Rolandos Liatsos 出演 Woman in Mind']
分析：根据问题和 SPO 提及，我们需要找到出版物 “Woman's Viewpoint” 的出版国家。SPO 'Woman's Viewpoint 在德克萨斯州创立' 包含了地理位置信息，可从这里推断出所在国籍
output：['Woman's Viewpoint 在德克萨斯州创立']

示例 2:
问题：哪位德国音乐家的手稿是《C大调长笛奏鸣曲，BWV 1033》的？
SPO 提及：实体[Flute Sonata in C major, BWV 1033] InHandOf 实体
SPO 候选项：['C大调长笛奏鸣曲，BWV 1033 归因于 Johann Sebastian Bach', 'C大调长笛奏鸣曲，BWV 1033 是为长笛或竖笛和低音连续演奏', 'C大调长笛奏鸣曲，BWV 1033 是一首四乐章的奏鸣曲']
分析：根据问题和 SPO 提及，我们需要找出谁持有《C大调长笛奏鸣曲，BWV 1033》的手稿。根据提供的 SPO 候选项，SPO "C大调长笛奏鸣曲，BWV 1033 归因于 Johann Sebastian Bach" 与问题相关。可以推断出《C大调长笛奏鸣曲，BWV 1033》的手稿应在 Johann Sebastian Bach 的手中。
output：['C大调长笛奏鸣曲，BWV 1033 归因于 Johann Sebastian Bach']

要求：
你是一名语言专家。你的任务是根据以下规则从给定的 SPO 候选项中选择正确的 SPO 文本来回答给定的问题。请确保它与 SPO 提及或问题匹配。

输出必须从 SPO 候选项中选择，并且必须与其内容保持一致，以列表格式呈现。
如果在 SPO 候选项中没有合适的答案，输出一个空列表。确保输出与问题或 SPO 提及高度匹配。
如果在 SPO 候选项中有多个正确答案，输出所有匹配的 SPO。

问题：$question
SPO 提及: $mention
SPO 候选项: $candis
output:
"""
    template_en = """Examples 1:
Question: Is the Woman's Viewpoint a British publication?
spo mention: Publication[Woman's Viewpoint] Nationality Country
spo candications: ['the woman s viewpoint ranFrom 1923 to 1927', 'the woman s viewpoint publishedBy florence m  sterling', 'the woman s viewpoint foundedIn 1923', 'the woman s viewpoint foundedIn texas', 'the woman s viewpoint was a woman s magazine', 'rolandos liatsos starredIn woman in mind', 'rolandos liatsos starredIn woman in mind']
Analysis: Based on the Question and the SPO mention, we need to find the country of publication for the publication "Woman's Perspective". According to the information provided in the SPO candidates, the SPO  'the woman s viewpoint foundedIn texas' is related to the spo mention.
output: ['the woman s viewpoint foundedIn texas']

Examples 2:
Question: Who is the German musician whose hand the manuscript for Flute Sonata in C major, BWV 1033 is in?
spo mention: Entity[Flute Sonata in C major, BWV 1033] InHandOf Entity
spo candications: ['flute sonata in c major  bwv 1033 isAttributedTo johann sebastian bach', 'flute sonata in c major  bwv 1033 isFor flute or recorder and basso continuo', 'flute sonata in c major  bwv 1033 is a sonata in 4 movements']
Analysis: Based on the Question and the SPO mention, we need to find out who holds the manuscript of "Flute Sonata in C major, BWV 1033". According to the information provided in the SPO candidates, the SPO "flute sonata in c major bwv 1033 isAttributedTo johann sebastian bach" is related to the spo mention. It can be inferred that the manuscript of "Flute Sonata in C major, BWV 1033" should be in the hands of Johann Sebastian Bach.
output: ['flute sonata in c major  bwv 1033 isAttributedTo johann sebastian bach']

Requirements:
You are a language expert. Your task is to select the correct SPO (subject predicate object) text from the given SPO candidates according to the following rules to answer the given question. Please ensure that it matches the SPO mention or answer question.
1. The output must be selected from the SPO candidates and must remain consistent with their content, presented in a list format.
2. If there is no suitable answer in the SPO candidates, output an empty list. Ensure that the output is highly matched with the question or SPO mention.
3. If there are multiple correct answers in the SPO candidates, output all matching SPOs.

Question：$question
spo mention: $mention
spo candications: $candis
output: """

    def __init__(self, language: str):
        super().__init__(language)

    @property
    def template_variables(self) -> List[str]:
        return ["question", "mention", "candis"]

    def parse_response_en(self, satisfied_info: str):
        if satisfied_info[:3] == 'Yes':
            if_finished = True
        else:
            if_finished = False
        return if_finished

    def parse_response_zh(self, satisfied_info: str):
        if satisfied_info.startswith("是"):
            if_finished = True
        else:
            if_finished = False
        return if_finished

    def parse_response(self, response: str, **kwargs):
        logger.debug(
            f"SpoRetrieval {response} mention:{self.template_variables_value.get('mention', '')} "
            f"candis:{self.template_variables_value.get('candis', '')}")
        llm_output = response.replace('Expected Output:', '')
        llm_output = llm_output.replace('"', '')
        return llm_output.strip()



import re
from string import Template
from typing import List
import logging

from kag.common.base.prompt_op import PromptOp

logger = logging.getLogger(__name__)


class RespVerifier(PromptOp):
    template_zh = "仅根据当前已知的信息，并且不允许进行推理，" \
                  "你能否完全并准确地回答这个问题'$sub_instruction'?\n已知信息：'$supporting_fact'。" \
                  "\n如果你能，请直接回复‘是’，并给出问题'$sub_instruction'的答案，无需重复问题；如果不可以，请直接回答'否'。"
    template_en = "Judging based solely on the current known information and without allowing for inference, " \
                 "are you able to respond completely and accurately to the question '$sub_instruction'? \n" \
                 "Known information: '$supporting_fact'. If yes, please reply with 'Yes', followed by an accurate response to the question '$sub_instruction', " \
                 "without restating the question; if no, please reply with 'No' directly."

    def __init__(self, language: str):
        super().__init__(language)


    @property
    def template_variables(self) -> List[str]:
        return ["sub_instruction", "supporting_fact"]

    def parse_response_en(self, satisfied_info: str):
        if satisfied_info[:3] == 'Yes':
            satisfied = True
        else:
            satisfied = False
        if satisfied:
            satisfied_info = satisfied_info.replace('Yes', '').strip()
            res = "The answer to the Question'{}' is '{}'".format(self.template_variables_value["sub_instruction"], satisfied_info)
            return res
        return None

    def parse_response_zh(self, satisfied_info: str):
        if satisfied_info.startswith("是"):
            satisfied = True
        else:
            satisfied = False
        if satisfied:
            satisfied_info = satisfied_info.replace('是', '').strip()
            res = "问题'{}' 的答案是 '{}'".format(self.template_variables_value["sub_instruction"], satisfied_info)
            return res
        return None

    def parse_response(self, response: str, **kwargs):
        logger.debug('推理器判别:{}'.format(response))
        if self.language == "en":
            return self.parse_response_en(response)
        return self.parse_response_zh(response)



from kag.solver.logic.core_modules.common.base_model import SPOEntity, LogicNode
from kag.solver.logic.core_modules.common.one_hop_graph import KgGraph, EntityData, RelationData
from kag.solver.logic.core_modules.common.schema_utils import SchemaUtils
from kag.solver.logic.core_modules.op_executor.op_executor import OpExecutor
from kag.solver.logic.core_modules.parser.logic_node_parser import GetNode
from kag.solver.logic.core_modules.retriver.entity_linker import spo_entity_linker, EntityLinkerBase
from kag.solver.logic.core_modules.retriver.graph_retriver.dsl_executor import DslRunner


class GetExecutor(OpExecutor):
    def __init__(self, nl_query: str, kg_graph: KgGraph, schema: SchemaUtils, el: EntityLinkerBase,
                 dsl_runner: DslRunner, cached_map: dict, debug_info: dict, **kwargs):
        super().__init__(nl_query, kg_graph, schema, debug_info, **kwargs)

        self.el = el
        self.dsl_runner = dsl_runner
        self.cached_map = cached_map

    def executor(self, logic_node: LogicNode, req_id: str, param: dict) -> list:
        kg_qa_result = []
        if not isinstance(logic_node, GetNode) or not self.debug_info.get('exact_match_spo', False):
            return kg_qa_result

        n = logic_node
        s_data_set = self.kg_graph.get_entity_by_alias(n.alias_name)
        if isinstance(n.s, SPOEntity) and s_data_set is None:
            if len(n.s.id_set) > 0:
                start_info_set = n.s.generate_start_infos()
                for start_info in start_info_set:
                    id_info = EntityData()
                    id_info.type = start_info['type']
                    id_info.biz_id = start_info['id']
                    s_data_set = [id_info]
            elif n.s.entity_name:
                el_results, el_request, err_msg, call_result_data = spo_entity_linker(self.kg_graph,
                                                                                      n,
                                                                                      self.nl_query,
                                                                                      self.el,
                                                                                      self.schema,
                                                                                      req_id,
                                                                                      param)
                self.debug_info['el'] = self.debug_info['el'] + el_results
                self.debug_info['el_detail'] = self.debug_info['el_detail'] + [{
                    "el_request": el_request,
                    'el_results': el_results,
                    'el_debug_result': call_result_data,
                    'err_msg': err_msg
                }]
                n.to_std(n.args)
                s_data_set = self.kg_graph.get_entity_by_alias(n.alias_name)
        if s_data_set is None:
            self.debug_info['get_empty'].append(n.to_dict())
            return kg_qa_result

        s_biz_id_set = []
        for s_data in s_data_set:
            if isinstance(s_data, EntityData):
                if s_data.name == '':
                    s_biz_id_set.append(s_data.biz_id)
                else:
                    kg_qa_result.append(s_data.name)
            if isinstance(s_data, RelationData):
                kg_qa_result.append(str(s_data))
        if len(s_biz_id_set) > 0:
            one_hop_cached_map = self.dsl_runner.query_vertex_property_by_s_ids(s_biz_id_set,
                                                                                n.s.get_entity_first_type(),
                                                                                self.cached_map)

            self.kg_graph.nodes_alias.append(n.alias_name)
            entities = []
            for one_hop in one_hop_cached_map.keys():
                kg_qa_result.append(one_hop_cached_map[one_hop].s.name)
                entities.append(one_hop_cached_map[one_hop].s)

            if n.alias_name not in self.kg_graph.entity_map.keys():
                self.kg_graph.entity_map[n.alias_name] = entities
            else:
                self.kg_graph.entity_map[n.alias_name] = self.kg_graph.entity_map[n.alias_name] + entities
        return kg_qa_result



from tenacity import retry, stop_after_attempt

from kag.common.base.prompt_op import PromptOp
from kag.interface.solver.kag_memory_abc import KagMemoryABC


class DefaultMemory(KagMemoryABC):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.verify_prompt = PromptOp.load(self.biz_scene, "resp_verifier")(
            language=self.language
        )
        self.extractor_prompt = PromptOp.load(self.biz_scene, "resp_extractor")(
            language=self.language
        )
        self.state_memory = []
        self.evidence_memory = []
        self.exact_answer = []
        self.instruction_set = []

    @retry(stop=stop_after_attempt(3))
    def _verifier(self, supporting_fact, sub_instruction):
        res = self.llm_module.invoke({'sub_instruction': sub_instruction,
                                      'supporting_fact': supporting_fact}, self.verify_prompt,
                                     with_json_parse=False, with_except=True)
        if res is None:
            return
        if res not in self.state_memory:
            self.state_memory.append(res)

    @retry(stop=stop_after_attempt(3))
    def _extractor(self, supporting_fact, instruction):
        if supporting_fact is None or supporting_fact == '':
            return
        evidence = self.llm_module.invoke({'supporting_fact': supporting_fact, 'instruction': instruction},
                                          self.extractor_prompt, with_json_parse=False, with_except=True)
        if evidence not in self.evidence_memory:
            self.evidence_memory.append(evidence)

    def save_memory(self, solved_answer, supporting_fact, instruction):
        if solved_answer != "":
            self.exact_answer.append(solved_answer)
            return
        # skip first instruction to verifier
        if len(self.instruction_set) != 0:
            self._verifier(supporting_fact, instruction)

        self._extractor(supporting_fact, instruction)
        self.instruction_set.append(instruction)

    def get_solved_answer(self):
        return self.exact_answer[-1] if len(self.exact_answer) > 0 else None

    def serialize_memory(self):
        if len(self.exact_answer) > 0:
            return f"[Solved Answer]{self.exact_answer[-1]}"
        serialize_memory = "[State Memory]:{}\n[Evidence Memory]:{}\n".format(
            str(self.state_memory), str(self.evidence_memory)
        )
        return serialize_memory

    def refresh(self):
        self.state_memory = []
        self.evidence_memory = []
        self.exact_answer = []


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
from collections import defaultdict
from typing import Dict, List, Callable

import pandas

from knext.schema.client import BASIC_TYPES
from kag.builder.model.sub_graph import SubGraph, Node
from knext.common.base.runnable import Input, Output
from knext.schema.client import SchemaClient
from knext.schema.model.base import SpgTypeEnum

from knext.schema.model.schema_helper import (
    SPGTypeName,
    PropertyName,
)
from kag.interface.builder.mapping_abc import MappingABC

FuseFunc = Callable[[SubGraph], List[SubGraph]]
LinkFunc = Callable[[str, Node], List[Node]]


class SPGTypeMapping(MappingABC):
    """
    A class for mapping SPG (Simple Property Graph) types and handling their properties and strategies.

    Attributes:
        spg_type_name (SPGTypeName): The name of the SPG type.
        fuse_op (FuseOpABC, optional): The user-defined fuse operator. Defaults to None.
    """

    def __init__(self, spg_type_name: SPGTypeName, fuse_func: FuseFunc = None, **kwargs):
        super().__init__(**kwargs)
        self.schema = SchemaClient(project_id=self.project_id).load()
        assert (
            spg_type_name in self.schema
        ), f"SPG type [{spg_type_name}] does not exist."
        self.spg_type = self.schema.get(spg_type_name)

        self.property_mapping: Dict = defaultdict(list)
        self.link_funcs: Dict = dict()
        self.fuse_func = fuse_func

    def add_property_mapping(
        self,
        source_name: str,
        target_name: PropertyName,
        link_func: LinkFunc = None,
    ):
        """
        Adds a property mapping from a source name to a target name within the SPG type.

        Args:
            source_name (str): The source name of the property.
            target_name (PropertyName): The target name of the property within the SPG type.
            link_func (LinkFunc, optional): The user-defined link operator. Defaults to None.

        Returns:
            self
        """
        if (
            target_name not in ["id", "name"]
            and target_name not in self.spg_type.properties
        ):
            raise ValueError(
                f"Property [{target_name}] does not exist in [{self.spg_type.name}]."
            )

        self.property_mapping[target_name].append(source_name)
        if link_func is not None:
            self.link_funcs[target_name] = link_func
        return self

    @property
    def input_types(self) -> Input:
        return Dict[str, str]

    @property
    def output_types(self) -> Output:
        return SubGraph

    def field_mapping(self, record: Dict[str, str]) -> Dict[str, str]:
        """
        Maps fields from a record based on the defined property mappings.

        Args:
            record (Dict[str, str]): The input record containing source names and values.

        Returns:
            Dict[str, str]: A mapped record with target names and corresponding values.
        """
        mapped_record = {}
        for target_name, source_names in self.property_mapping.items():
            for source_name in source_names:
                value = record.get(source_name)
                mapped_record[target_name] = value
        return mapped_record

    def assemble_sub_graph(self, properties: Dict[str, str]):
        """
        Assembles a sub-graph based on the provided properties and linking strategies.

        Args:
            properties (Dict[str, str]): The properties to be used for assembling the sub-graph.

        Returns:
            SubGraph: The assembled sub-graph.
        """
        sub_graph = SubGraph(nodes=[], edges=[])
        s_id = properties.get("id", "")
        s_name = properties.get("name", s_id)
        s_label = self.spg_type.name_en

        for prop_name, prop_value in properties.items():
            if not prop_value or prop_value == pandas.NaT:
                continue
            if prop_name in self.spg_type.properties:
                prop = self.spg_type.properties.get(prop_name)
                o_label = prop.object_type_name_en
                if o_label not in BASIC_TYPES:
                    prop_value_list = prop_value.split(",")
                    for o_id in prop_value_list:
                        if prop_name in self.link_funcs:
                            link_func = self.link_funcs.get(prop_name)
                            o_ids = link_func(o_id, properties)
                            for _o_id in o_ids:
                                sub_graph.add_edge(
                                    s_id=s_id,
                                    s_label=s_label,
                                    p=prop_name,
                                    o_id=_o_id,
                                    o_label=o_label,
                                )
                        else:
                            sub_graph.add_edge(
                                s_id=s_id,
                                s_label=s_label,
                                p=prop_name,
                                o_id=o_id,
                                o_label=o_label,
                            )
        if self.spg_type.spg_type_enum == SpgTypeEnum.Concept:
            self.hypernym_predicate(sub_graph, s_id)
        else:
            sub_graph.add_node(
                id=s_id, name=s_name, label=s_label, properties=properties
            )

        return sub_graph

    def hypernym_predicate(self, sub_graph: SubGraph, concept_id: str):
        """
        Adds hypernym predicates to the sub-graph based on the provided concept ID.

        Args:
            sub_graph (SubGraph): The sub-graph to which hypernym predicates will be added.
            concept_id (str): The ID of the concept.
        """
        p = getattr(self.spg_type, "hypernym_predicate") or "isA"
        label = self.spg_type.name_en
        concept_list = concept_id.split("-")

        father_id = ""
        for concept_name in concept_list:
            concept_id = father_id + "-" + concept_name if father_id else concept_name
            sub_graph.add_node(id=concept_id, name=concept_name, label=label)
            if father_id:
                sub_graph.add_edge(
                    s_id=concept_id, s_label=label, p=p, o_id=father_id, o_label=label
                )
            father_id = concept_id

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Invokes the mapping process on the given input and returns the resulting sub-graphs.

        Args:
            input (Input): The input data to be processed.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list of resulting sub-graphs.
        """
        if self.property_mapping:
            properties = self.field_mapping(input)
        else:
            properties = input
        sub_graph = self.assemble_sub_graph(properties)
        return [sub_graph]



import json
import logging
from enum import Enum

from knext.common.rest import ApiClient, Configuration
from knext.reasoner.rest.models.ca_pipeline import CaPipeline
from knext.reasoner.rest.models.edge import Edge
from knext.reasoner.rest.models.node import Node
from knext.reasoner.rest.models.report_pipeline_request import ReportPipelineRequest
from knext.reasoner.rest.reasoner_api import ReasonerApi

logger = logging.getLogger(__name__)
class ReporterIntermediateProcessTool:
    class STATE(str, Enum):
        WAITING = "WAITING"
        RUNNING = "RUNNING"
        FINISH = "FINISH"
        ERROR = "ERROR"

    ROOT_ID = 0

    def __init__(self, report_log=False, task_id=None, project_id=None, host_addr=None):
        self.report_log = report_log
        self.task_id = task_id
        self.project_id = project_id
        self.client: ReasonerApi = ReasonerApi(
            api_client=ApiClient(configuration=Configuration(host=host_addr)))

    def report_pipeline(self, question, rewrite_question_list=[]):
        # print(question)
        for idx, item in enumerate(rewrite_question_list, start=2):
            item.id = idx
            # print(item)

        pipeline = CaPipeline()
        pipeline.nodes = []
        pipeline.edges = []
        pipeline.nodes.append(Node(id=self.ROOT_ID, state=self.STATE.WAITING, question=question.question, answer=None, logs=None))
        dep_question_list = []
        for item in rewrite_question_list:
            pipeline.nodes.append(Node(id=item.id, state=self.STATE.WAITING, question=item.question, answer=None, logs=None))
            if item.dependencies:
                for dep_item in item.dependencies:
                    pipeline.edges.append(Edge(_from=dep_item.id, to=item.id))
                    dep_question_list.append(dep_item)
        for item in rewrite_question_list:
            if item not in dep_question_list:
                pipeline.edges.append(Edge(_from=item.id, to=self.ROOT_ID))
        to_list = []
        for edge in pipeline.edges:
            to_list.append(edge.to)
        first_nodes = []
        for node in pipeline.nodes:
            if node.id not in to_list:
                first_nodes.append(node.id)
        # str([n.question for n in pipeline.nodes if n.id != self.ROOT_ID])
        pipeline.nodes.insert(0, Node(id=1, state=self.STATE.FINISH, question=question.question, answer=str([n.question for n in pipeline.nodes if n.id != self.ROOT_ID]), logs=None))
        for n in first_nodes:
            pipeline.edges.insert(0, Edge(_from=1, to=n))
        request = ReportPipelineRequest(task_id=self.task_id, pipeline=pipeline)
        if self.report_log:
            self.client.reasoner_dialog_report_pipeline_post(report_pipeline_request=request)
        else:
            logger.info(request)

    def report_node(self, question, answer, state):
        logs = self.format_logs(question.context)
        if not question.id:
            question.id = self.ROOT_ID
        node = Node(id=(question.id+1 if question.id != 0 else 0), state=state, question=question.question, answer=answer,
        logs=logs)
        request = ReportPipelineRequest(task_id=self.task_id, node=node)
        if self.report_log:
            self.client.reasoner_dialog_report_node_post(report_pipeline_request=request)
        else:
            logger.info(request)

    def format_logs(self, logs):
        if not logs:
            return None
        content = ""
        if isinstance(logs, str):
            try:
                logs = json.loads(logs)
            except:
                logs = eval(logs)
        for idx, item in enumerate(logs, start=1):
            content += f"{item}\n"
        return content



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import os

from kag.common.llm.client import LLMClient
from kag.solver.common.base import Question
from kag.solver.implementation.default_kg_retrieval import KGRetrieverByLlm
from kag.solver.implementation.default_lf_planner import DefaultLFPlanner
from kag.solver.implementation.default_reasoner import DefaultReasoner
from kag.solver.implementation.lf_chunk_retriever import LFChunkRetriever
from kag.solver.logic.core_modules.lf_solver import LFSolver
from kag.solver.logic.solver_pipeline import SolverPipeline
from kag.solver.tools.info_processor import ReporterIntermediateProcessTool


class SolverMain:

    def invoke(self, project_id: int, task_id: int, query: str, report_tool=True, host_addr="http://127.0.0.1:8887"):
        # resp
        report_tool = ReporterIntermediateProcessTool(report_log=report_tool, task_id=task_id, project_id=project_id, host_addr=host_addr)

        lf_planner = DefaultLFPlanner(KAG_PROJECT_ID=project_id, KAG_PROJECT_HOST_ADDR=host_addr)
        lf_solver = LFSolver(
            kg_retriever=KGRetrieverByLlm(KAG_PROJECT_ID=project_id, KAG_PROJECT_HOST_ADDR=host_addr),
            chunk_retriever=LFChunkRetriever(KAG_PROJECT_ID=project_id, KAG_PROJECT_HOST_ADDR=host_addr),
            report_tool=report_tool,
            KAG_PROJECT_ID=project_id,
            KAG_PROJECT_HOST_ADDR=host_addr
        )
        reason = DefaultReasoner(lf_planner=lf_planner, lf_solver=lf_solver, KAG_PROJECT_ID=project_id, KAG_PROJECT_HOST_ADDR=host_addr)
        question = Question(query)
        question.id = 0
        resp = SolverPipeline(reasoner=reason, KAG_PROJECT_ID=project_id, KAG_PROJECT_HOST_ADDR=host_addr)
        answer, trace_log = resp.run(query)
        print(trace_log)
        report_tool.report_node(question, answer, ReporterIntermediateProcessTool.STATE.FINISH)
        return answer


if __name__ == "__main__":
    res = SolverMain().invoke(3, 283, "周杰伦在哪一年基于什么作品获得的全球畅销专辑榜”冠军的华语歌手", True, host_addr="http://127.0.0.1:8887")
    print("*" * 80)
    print("The Answer is: ", res)



# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

import json
from kag.common.llm.client import LLMClient


class LLMConfigChecker(object):
    """
    Check whether the llm config is valid.
    """

    def check(self, config: str) -> str:
        """
        Check the llm config.

        * If the config is valid, return the generated text.

        * If the config is invalid, raise a RuntimeError exception.

        :param config: llm config
        :type config: str
        :return: the generated text
        :rtype: str
        :raises RuntimeError: if the config is invalid
        """
        config = json.loads(config)
        llm_client = LLMClient.from_config(config)
        try:
            res = llm_client("who are you?")
            return res
        except Exception as ex:
            raise RuntimeError(f"invalid llm config: {config}, for details: {ex}")
        
if __name__ == "__main__":
    config = '''
        {"client_type" :"ollama",
        "base_url" : "http://localhost:11434/",
        "model" : "llama3.1" }
    '''
    config_checker = LLMConfigChecker()
    res = config_checker.check(config)



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.
import hashlib
from enum import Enum
from typing import Dict, Any


class ChunkTypeEnum(str, Enum):
    Table = "TABLE"
    Text = "TEXT"


class Chunk:
    def __init__(
        self,
        id: str,
        name: str,
        content: str,
        type: ChunkTypeEnum = ChunkTypeEnum.Text,
        **kwargs
    ):
        self.id = id
        self.name = name
        self.type = type
        self.content = content
        self.kwargs = kwargs

    @staticmethod
    def generate_hash_id(value):
        if isinstance(value, str):
            value = value.encode("utf-8")
        hasher = hashlib.sha256()
        hasher.update(value)
        return hasher.hexdigest()

    def __str__(self):
        tmp = {
            "id": self.id,
            "name": self.name,
            "content": self.content
            if len(self.content) <= 64
            else self.content[:64] + " ...",
        }
        return f"<Chunk>: {tmp}"

    __repr__ = __str__

    def to_dict(self):
        return {
            "id": self.id,
            "name": self.name,
            "content": self.content,
            "type": self.type.value if isinstance(self.type, ChunkTypeEnum) else self.type,
            "properties": self.kwargs,
        }

    @classmethod
    def from_dict(cls, input_: Dict[str, Any]):
        return cls(
            id=input_.get("id"),
            name=input_.get("name"),
            content=input_.get("content"),
            type=input_.get("type"),
            **input_.get("properties", {}),
        )



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from typing import Dict, List

from knext.common.base.runnable import Input, Output
from kag.interface.builder import ExtractorABC


class UserDefinedExtractor(ExtractorABC):
    @property
    def input_types(self) -> Input:
        return Dict[str, str]

    @property
    def output_types(self) -> Output:
        return Dict[str, str]

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        return input



# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.

from collections import defaultdict
from typing import Dict, List

from kag.builder.model.sub_graph import SubGraph
from knext.common.base.runnable import Input, Output
from knext.schema.client import SchemaClient

from knext.schema.model.schema_helper import (
    SPGTypeName,
    RelationName,
)
from kag.interface.builder.mapping_abc import MappingABC


class RelationMapping(MappingABC):
    """
    A class that handles relation mappings by assembling subgraphs based on given subject, predicate, and object names.
    This class extends the Mapping class.

    Args:
        subject_name (SPGTypeName): The name of the subject type.
        predicate_name (RelationName): The name of the predicate.
        object_name (SPGTypeName): The name of the object type.
    """

    def __init__(
        self,
        subject_name: SPGTypeName,
        predicate_name: RelationName,
        object_name: SPGTypeName,
        **kwargs
    ):
        super().__init__(**kwargs)
        schema = SchemaClient(project_id=self.project_id).load()
        assert subject_name in schema, f"{subject_name} is not a valid SPG type name"
        assert object_name in schema, f"{object_name} is not a valid SPG type name"
        self.subject_type = schema.get(subject_name)
        self.object_type = schema.get(object_name)

        assert predicate_name in self.subject_type.properties or predicate_name in set(
            [key.split("_")[0] for key in self.subject_type.relations.keys()]
        ), f"{predicate_name} is not a valid SPG property/relation name"
        self.predicate_name = predicate_name

        self.src_id_field = None
        self.dst_id_field = None
        self.property_mapping: Dict = defaultdict(list)
        self.linking_strategies: Dict = dict()

    def add_src_id_mapping(self, source_name: str):
        """
        Adds a field mapping from source data to the subject's ID property.

        Args:
            source_name (str): The name of the source field to map.

        Returns:
            self
        """
        self.src_id_field = source_name
        return self

    def add_dst_id_mapping(self, source_name: str):
        """
        Adds a field mapping from source data to the object's ID property.

        Args:
            source_name (str): The name of the source field to map.

        Returns:
            self
        """
        self.dst_id_field = source_name
        return self

    def add_sub_property_mapping(self, source_name: str, target_name: str):
        """
        Adds a field mapping from source data to a property of the subject type.

        Args:
            source_name (str): The source field to be mapped.
            target_name (str): The target field to map the source field to.

        Returns:
            self
        """
        self.property_mapping[target_name].append(source_name)
        return self

    @property
    def input_types(self) -> Input:
        return Dict[str, str]

    @property
    def output_types(self) -> Output:
        return SubGraph

    def assemble_sub_graph(self, record: Dict[str, str]) -> SubGraph:
        """
        Assembles a subgraph from the provided record.

        Args:
            record (Dict[str, str]): The record containing the data to assemble into a subgraph.

        Returns:
            SubGraph: The assembled subgraph.
        """
        sub_graph = SubGraph(nodes=[], edges=[])
        if self.property_mapping:
            s_id = record.get(self.src_id_field or "srcId")
            o_id = record.get(self.dst_id_field or "dstId")
            sub_properties = {}
            for target_name, source_names in self.property_mapping.items():
                for source_name in source_names:
                    value = record.get(source_name)
                    sub_properties[target_name] = value
        else:
            s_id = record.pop(self.src_id_field or "srcId")
            o_id = record.pop(self.dst_id_field or "dstId")
            sub_properties = record

        sub_graph.add_edge(
            s_id=s_id,
            s_label=self.subject_type.name_en,
            p=self.predicate_name,
            o_id=o_id,
            o_label=self.object_type.name_en,
            properties=sub_properties,
        )

        return sub_graph

    def invoke(self, input: Input, **kwargs) -> List[Output]:
        """
        Invokes the assembly process to create a subgraph from the input data.

        Args:
            input (Input): The input data to assemble into a subgraph.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Output]: A list containing the assembled subgraph.
        """
        sub_graph = self.assemble_sub_graph(input)
        return [sub_graph]



from pydantic import Field
from kag.common.llm.config.base import LLMConfig


class OpenAIConfig(LLMConfig):
    api_key: str = Field(
        description="api key."
    )
    stream: bool = Field(
        description="if use stream mode",default=False
    )
    model: str = Field(
        description="model name."
    )
    temperature: float = Field(
        description="temperature.",default=0.7
    )
    base_url: str = Field(
        description="post url."
    )


# -*- coding: utf-8 -*-
# Copyright 2023 OpenSPG Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
# in compliance with the License. You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License
# is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied.


import json
from typing import Union
from openai import OpenAI
import logging

from kag.common.llm.client.llm_client import LLMClient
from kag.common.llm.config import OpenAIConfig

# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


class OpenAIClient(LLMClient):
    """
    A client class for interacting with the OpenAI API.

    Initializes the client with an API key, base URL, streaming option, temperature parameter, and default model.

    Parameters:
        api_key (str): The OpenAI API key.
        base_url (str): The base URL of the API.
        stream (bool, optional): Whether to process responses in a streaming manner. Default is False.
        temperature (int, optional): Sampling temperature to control the randomness of the model's output. Default is 0.7.
        model (str, optional): The default model to use.

    Attributes:
        api_key (str): The OpenAI API key.
        base_url (str): The base URL of the API.
        model (str): The default model to use.
        stream (bool): Whether to process responses in a streaming manner.
        temperature (float): Sampling temperature.
        client (OpenAI): An instance of the OpenAI API client.
    """
    def __init__(
            self,
            llm_config:OpenAIConfig
    ):
        # Initialize the OpenAIClient object
        self.api_key = llm_config.api_key
        self.base_url = llm_config.base_url
        self.model = llm_config.model
        self.stream = llm_config.stream
        self.temperature = llm_config.temperature
        self.client = OpenAI(api_key=self.api_key, base_url=self.base_url)


    def __call__(self, prompt:str, image_url:str=None):
        """
        Executes a model request when the object is called and returns the result.

        Parameters:
            prompt (str): The prompt provided to the model.

        Returns:
            str: The response content generated by the model.
        """
        # Call the model with the given prompt and return the response
        if image_url:
            message = [
                {"role": "system", "content": "you are a helpful assistant"},
                {"role": "user", "content": [
                    {
                        "type": "text",
                        "text": prompt
                    },
                    {
                        "type": "image_url",
                        "image_url": {
                        "url": image_url
                        }
                    }
                    ]
                },
            ]
            response = self.client.chat.completions.create(
                model=self.model,
                messages=message,
                stream=self.stream,
                temperature=self.temperature,
            )
            rsp = response.choices[0].message.content
            return rsp
            
        else:
            message = [
                {"role": "system", "content": "you are a helpful assistant"},
                {"role": "user", "content": prompt},
            ]
            response = self.client.chat.completions.create(
                model=self.model,
                messages=message,
                stream=self.stream,
                temperature=self.temperature,
            )
            rsp = response.choices[0].message.content
            return rsp

    def call_with_json_parse(self, prompt):
        """
        Calls the model and attempts to parse the response into JSON format.

        Parameters:
            prompt (str): The prompt provided to the model.

        Returns:
            Union[dict, str]: If the response is valid JSON, returns the parsed dictionary; otherwise, returns the original response.
        """
        # Call the model and attempt to parse the response into JSON format
        rsp = self(prompt)
        _end = rsp.rfind("```")
        _start = rsp.find("```json")
        if _end != -1 and _start != -1:
            json_str = rsp[_start + len("```json"): _end].strip()
        else:
            json_str = rsp
        try:
            json_result = json.loads(json_str)
        except:
            return rsp
        return json_result


from abc import ABC, abstractmethod

from kag.interface.solver.kag_memory_abc import KagMemoryABC
from kag.solver.common.base import KagBaseModule


class KAGGeneratorABC(KagBaseModule, ABC):
    """
     The Generator class is an abstract base class for generating responses using a language model module.
     It initializes prompts for judging and generating responses based on the business scene and language settings.
     """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def generate(self, instruction, memory: KagMemoryABC) -> str:
        raise NotImplementedError("Subclasses must implement this method")
